import os
from .ffterm import *
from .errors import *
from mstk import MSTK_FORCEFIELD_PATH, logger


class ForceField:
    '''
    ForceField is a set of :class:`FFTerm` objects describing the energy function of
    a :class:`~mstk.topology.Topology` with regard to the positions of atoms.

    Attributes
    ----------
    atom_types : dict, [str, AtomType]
        All the atom types in this force field.
        For each key value pair, the value is a AtomType object,
        the key is the :attr:`~AtomType.name` of the object.
    qinc_terms : dict, [str, ChargeIncrementTerm]
        All the charge increment terms.
        For each key value pair, the value is a ChargeIncrementTerm object,
        the key is the :attr:`~ChargeIncrementTerm.name` of the object.
    vdw_terms : dict, [str, subclass of VdwTerm]
        vdW terms between same types.
        For each key value pair, the value is a object of subclass of VdwTerm.
        the key is the :attr:`~VdwTerm.name` of the object.
        The term can be either :class:`LJ126Term` or :class:`MieTerm`,
        as long as it's a subclass of VdwTerm.
        Because we are using name of the term as key,
        there can only be one term to describe the vdW interactions between two atom types.
        Which means, there cannot be a LJ126Term and a MieTerm with same name in a ForceField.
        However, you can have LJ126Term for one pair of atom types and MieTerm for another pairs.
    pairwise_vdw_terms : dict, [str, subclass of VdwTerm]
        Pairwise vdW terms between unlike types.
        For each key value pair, the value is a object of subclass of VdwTerm.
        the key is the :attr:`~VdwTerm.name` of the object.
        If a pair of unlike atom types is not included in this dict,
        then the vdW interactions between this pair will be generated from mixing rule.
    polarizable_terms : dict, [str, subclass of PolarizableTerm]
        Polarization terms for polarizable atom types.
        For each key value pair, the value is a object of subclass of PolarizableTerm.
        the key is the :attr:`~PolarizableTerm.name` of the object.
        Currently, isotropic Drude model is implemented.
    virtual_site_terms: dict, [str, subclass of VirtualSiteTerm]
        Virtual site terms for delocalized interaction center.
        For each key value pair, the value is a object of subclass of VirtualSiteTerm.
        the key is the :attr:`~VirtualSiteTerm.name` of the object.
        Currently, TIP4P model is implemented.
    bond_terms : dict, [str, subclass of BondTerm]
        Bond terms between two atom types.
        For each key value pair, the value is a object of subclass of BondTerm.
        the key is the :attr:`~BondTerm.name` of the object.
    angle_terms : dict, [str, subclass of AngleTerm]
        Angle terms between three atom types.
        For each key value pair, the value is a object of subclass of AngleTerm.
        the key is the :attr:`~AngleTerm.name` of the object.
    dihedral_terms : dict, [str, subclass of DihedralTerm]
        Dihedral terms between four atom types.
        For each key value pair, the value is a object of subclass of DihedralTerm.
        the key is the :attr:`~DihedralTerm.name` of the object.
    improper_terms : dict, [str, subclass of ImproperTerm]
        Improper terms between four atom types.
        For each key value pair, the value is a object of subclass of ImproperTerm.
        the key is the :attr:`~ImproperTerm.name` of the object.
    vdw_cutoff : float
        Cutoff distance for vdW interactions.
        Will be ignored if periodic boundary condition is not used.
    vdw_long_range : str
        The method to handle the long-range vdW interactions.
        Optional values: :attr:`VDW_LONGRANGE_CORRECT`, :attr:`VDW_LONGRANGE_SHIFT`
        Will be ignored if periodic boundary condition is not used.
    lj_mixing_rule : str
        The rule to generate the pairwise LJ126 parameters if not provided.
        Optional values: :attr:`LJ_MIXING_NONE`, :attr:`LJ_MIXING_LB`, :attr:`LJ_MIXING_GEOMETRIC`
    scale_14_vdw : float
        The scaling factor for 1-4 vdW interactions
    scale_14_coulomb : float
        The scaling factor for 1-4 Coulomb interactions
    comments : list of str
        Comments describes misc information for this force field
    '''
    #: Truncated vdW interactions with long range continuum correction
    VDW_LONGRANGE_CORRECT = 'correct'
    #: Truncated vdW interactions with potential shifted so that vdW energy equal to zero at cutoff
    VDW_LONGRANGE_SHIFT = 'shift'

    #: Do not use mixing rule for LJ126 parameters
    LJ_MIXING_NONE = 'none'
    #: Lorentz-Berthelot mixing rule for LJ126 parameters (used by CHARMM and AMBER)
    LJ_MIXING_LB = 'lorentz-berthelot'
    #: Geometric mixing rule for LJ126 parameters (used by OPLS)
    LJ_MIXING_GEOMETRIC = 'geometric'

    SETTING_ATTRS = {'vdw_cutoff'      : float,
                     'vdw_long_range'  : str,
                     'lj_mixing_rule'  : str,
                     'scale_14_vdw'    : float,
                     'scale_14_coulomb': float,
                     }

    _class_map = {}

    @staticmethod
    def registor_format(extension, cls):
        ForceField._class_map[extension] = cls

    def __init__(self):
        self.atom_types: {str: AtomType} = {}
        self.qinc_terms: {str: ChargeIncrementTerm} = {}
        self.bond_terms: {str: BondTerm} = {}
        self.angle_terms: {str: AngleTerm} = {}
        self.dihedral_terms: {str: DihedralTerm} = {}
        self.improper_terms: {str: ImproperTerm} = {}
        self.vdw_terms: {str: VdwTerm} = {}
        self.pairwise_vdw_terms: {str: VdwTerm} = {}
        self.polarizable_terms: {str: PolarizableTerm} = {}
        self.virtual_site_terms: {str: VirtualSiteTerm} = {}

        self.vdw_cutoff = 1.0  # nm
        self._vdw_long_range = ForceField.VDW_LONGRANGE_CORRECT
        self._lj_mixing_rule = ForceField.LJ_MIXING_GEOMETRIC
        self.scale_14_vdw = 0.5
        self.scale_14_coulomb = 0.5

        self.comments: [str] = []

    @property
    def vdw_long_range(self):
        return self._vdw_long_range

    @vdw_long_range.setter
    def vdw_long_range(self, val):
        if val not in [ForceField.VDW_LONGRANGE_CORRECT, ForceField.VDW_LONGRANGE_SHIFT]:
            raise Exception(f'Invalid vdw_long_range: {val}')
        self._vdw_long_range = val

    @property
    def lj_mixing_rule(self):
        return self._lj_mixing_rule

    @lj_mixing_rule.setter
    def lj_mixing_rule(self, val):
        if val not in [ForceField.LJ_MIXING_NONE, ForceField.LJ_MIXING_LB, ForceField.LJ_MIXING_GEOMETRIC]:
            raise Exception(f'Invalid lj_mixing_rule: {val}')
        self._lj_mixing_rule = val

    @staticmethod
    def open(*files):
        '''
        Load ForceField from one or multiple files.

        If a file does not exist, will search it under directories defined by `MSTK_FORCEFIELD_PATH`.
        The parser will be determined from the extension of each file.

        If multiple files are provided, they will be parsed in order and be combined into a single ForceField object.
        These force fields must share the same setting. Otherwise an Exception will be raised.
        If the force fields contain comments, only the ones from the first file will be kept.
        If the same parameters appeared in several files, the ones parsed later will override the ones parsed earlier.

        Parameters
        ----------
        files : str
            The files to be read. The extension can be `zfp`, `zff', `ppf` or `ff`.

            * `zfp` file is the native XML format of mstk, and will be parsed by :class:`Zfp`.
            * `zff` file is the plain text format of mstk, and will be parsed by :class:`Zff`.
            * `ppf` file will be parsed as DFF format by :class:`Ppf`.
            * `ff` file will be parsed as fftool format by :class:`Padua`.

        Returns
        -------
        forcefield : ForceField

        '''
        files_path = []
        for i, file in enumerate(files):
            for dir in ['.'] + MSTK_FORCEFIELD_PATH:
                p = os.path.join(dir, file)
                if os.path.exists(p):
                    files_path.append(p)
                    break
            else:
                raise Exception(f'FF file not found: {file}')

        ffs = []
        for path in files_path:
            basename, ext = os.path.splitext(path)
            try:
                cls = ForceField._class_map[ext]
            except KeyError:
                raise Exception(f'Unsupported FF file format: {ext}')
            ffs.append(cls(path).forcefield)

        ff0: ForceField = ffs[0]
        if len(ffs) > 1:
            for ffi in ffs[1:]:
                ffi: ForceField
                if ff0.get_settings() != ffi.get_settings():
                    raise Exception('Settings for several FFs are not the same')
                for term in ffi.get_all_terms():
                    ff0.add_term(term, replace=True)
        return ff0

    def write(self, file):
        '''
        Write this force field into a file.

        The format of the file will be determined by its extension

        Parameters
        ----------
        file : str
        '''
        basename, ext = os.path.splitext(file)
        try:
            cls = ForceField._class_map[ext]
        except KeyError:
            raise Exception(f'Unsupported FF format: {file}')

        cls.save_to(self, file)

    @property
    def is_polarizable(self):
        '''
        Whether or not this is a polarizable force field.

        Return True as long as the :attr:`polarizable_terms` is not empty.

        Returns
        -------
        is : bool
        '''
        return len(self.polarizable_terms) != 0

    @property
    def has_virtual_site(self):
        '''
        Whether or not this force field contains virtual sites.

        Return True as long as the :attr:`virtual_site_terms` is not empty.

        Returns
        -------
        is : bool
        '''
        return len(self.virtual_site_terms) != 0

    def get_settings(self):
        '''
        Get the settings of this force field.

        Returns
        -------
        settings : dict, [str, str]

        '''
        d = {}
        for i in ('vdw_cutoff', 'vdw_long_range', 'lj_mixing_rule',
                  'scale_14_vdw', 'scale_14_coulomb'):
            d[i] = getattr(self, i)
        return d

    def restore_settings(self, d):
        '''
        Restore the settings of this force field

        Parameters
        ----------
        d : dict, [str, str]

        '''
        for i in ('vdw_cutoff', 'vdw_long_range', 'lj_mixing_rule',
                  'scale_14_vdw', 'scale_14_coulomb'):
            setattr(self, i, d[i])

    def add_term(self, term, replace=False):
        '''
        Add a force field term into this force field.

        If a terms already exists in the force field,
        the argument `replace` will determine the behavior.
        If replace is True, the one existed will be replaced by the new term.
        Otherwise, an Exception will be raised.

        Parameters
        ----------
        term : subclass of FFTerm
        replace : bool
        '''
        _duplicated = False
        if isinstance(term, AtomType):
            if term.name not in self.atom_types or replace:
                self.atom_types[term.name] = term
            else:
                _duplicated = True
        elif isinstance(term, ChargeIncrementTerm):
            if term.name not in self.qinc_terms or replace:
                self.qinc_terms[term.name] = term
            else:
                _duplicated = True
        elif isinstance(term, BondTerm):
            if term.name not in self.bond_terms or replace:
                self.bond_terms[term.name] = term
            else:
                _duplicated = True
        elif isinstance(term, AngleTerm):
            if term.name not in self.angle_terms or replace:
                self.angle_terms[term.name] = term
            else:
                _duplicated = True
        elif isinstance(term, DihedralTerm):
            if term.name not in self.dihedral_terms or replace:
                self.dihedral_terms[term.name] = term
            else:
                _duplicated = True
        elif isinstance(term, ImproperTerm):
            if term.name not in self.improper_terms or replace:
                self.improper_terms[term.name] = term
            else:
                _duplicated = True
        elif isinstance(term, VdwTerm):
            if term.type1 == term.type2:
                if term.name not in self.vdw_terms or replace:
                    self.vdw_terms[term.name] = term
                else:
                    _duplicated = True
            else:
                if term.name not in self.pairwise_vdw_terms or replace:
                    self.pairwise_vdw_terms[term.name] = term
                else:
                    _duplicated = True
        elif isinstance(term, PolarizableTerm):
            if term.name not in self.polarizable_terms or replace:
                self.polarizable_terms[term.name] = term
            else:
                _duplicated = True
        elif isinstance(term, VirtualSiteTerm):
            if term.name not in self.virtual_site_terms or replace:
                self.virtual_site_terms[term.name] = term
            else:
                _duplicated = True
        else:
            raise Exception('Invalid term to add')

        if _duplicated:
            raise Exception(f'{str(term)} already exist in FF')

    def get_vdw_term(self, type1, type2, mixing=True):
        '''
        Get the vdW term between two atom types.

        If type1 and type2 are the same, then will search parameters in :attr:`vdw_terms` and return the reference of the term.
        If not found, an Exception will be raised.
        If type1 and type2 are not the same, then will search parameters in :attr:`pairwise_vdw_terms` and return the reference of the term.
        If not found and `mixing` is True and :attr:`lj_mixing_rule` does not equal to :attr:`LJ_MIXING_NONE`,
        then a vdw term will be constructed from the mixing rule and be returned.
        Otherwise, an Exception will be raised.
        The constructed vdw term will be :class:`LJ126Term` if vdw for both type1 and type2 are :class:`LJ126Term`,
        otherwise a :class:`MieTerm` will be constructed.

        Parameters
        ----------
        type1 : AtomType or str
            If type1 is a AtomType object, then will search for its equivalent type for vdW parameters.
            If type1 is a str, then will match exactly this type.
        type2 : AtomType or str
            If type2 is a AtomType object, then will search for its equivalent type for vdW parameters.
            If type2 is a str, them will match exactly this type.
        mixing : bool
            Whether to use mixing rule to construct a pairwise vdW term if not found from :attr:`pairwise_vdw_terms`.
            If set to False, and the pairwise vdW term not found, an Exception will be raised even if this force field supports mixing rule.

        Returns
        -------
        term : subclass of VdwTerm
        '''
        at1: str = type1.eqt_vdw if type(type1) is AtomType else type1
        at2: str = type2.eqt_vdw if type(type2) is AtomType else type2
        vdw = VdwTerm(at1, at2)
        if at1 == at2:
            vdw = self.vdw_terms.get(vdw.name)
            if vdw is not None:
                return vdw
            else:
                raise FFTermNotFoundError(f'VdwTerm for {at1} not found')

        vdw = self.pairwise_vdw_terms.get(vdw.name)

        if vdw is not None:
            return vdw
        elif not mixing:
            raise FFTermNotFoundError(f'VdwTerm for {at1}-{at2} not found')

        lj1 = self.vdw_terms.get(VdwTerm(at1, at1).name)
        lj2 = self.vdw_terms.get(VdwTerm(at2, at2).name)
        if lj1 is None or lj2 is None:
            raise FFTermNotFoundError(f'VdwTerm for {at1} or {at2} not found')
        if type(lj1) not in (LJ126Term, MieTerm) or type(lj2) not in (LJ126Term, MieTerm):
            raise Exception('Cannot use mixing rule for vdW terms other than LJ126Term and MieTerm')

        if self.lj_mixing_rule == self.LJ_MIXING_LB:
            epsilon = math.sqrt(lj1.epsilon * lj2.epsilon)
            sigma = (lj1.sigma + lj2.sigma) / 2
        elif self.lj_mixing_rule == self.LJ_MIXING_GEOMETRIC:
            epsilon = math.sqrt(lj1.epsilon * lj2.epsilon)
            sigma = math.sqrt(lj1.sigma * lj2.sigma)
        else:
            raise Exception('Unknown mixing rule for vdW terms')

        if type(lj1) is LJ126Term and type(lj2) is LJ126Term:
            return LJ126Term(at1, at2, epsilon, sigma)

        repulsion1, attraction1 = (lj1.repulsion, lj1.attraction) if type(lj1) is MieTerm else (12, 6)
        repulsion2, attraction2 = (lj2.repulsion, lj2.attraction) if type(lj2) is MieTerm else (12, 6)
        repulsion = math.sqrt(repulsion1 * repulsion2)
        attraction = math.sqrt(attraction1 * attraction2)
        return MieTerm(at1, at2, epsilon, sigma, repulsion, attraction)

    def get_eqt_for_bond(self, bond):
        '''
        Get the equivalent atom types for bond parameters of the two atoms in a bond.

        Parameters
        ----------
        bond : ~mstk.topology.Bond

        Returns
        -------
        type1 : str
            The equivalent atom type for bond parameters of the first atom in this bond.
        type2 : str
            The equivalent atom type for bond parameters of the second atom in this bond.
        '''
        try:
            at1 = self.atom_types.get(bond.atom1.type).eqt_bond
            at2 = self.atom_types.get(bond.atom2.type).eqt_bond
        except:
            raise FFTermNotFoundError(
                f'Atom type {bond.atom1.type} or {bond.atom2.type} not found in FF')
        return at1, at2

    def get_bond_term(self, type1, type2):
        '''
        Get the bond term for the bond formed by two atom types.

        If not found, an Exception will be raised.

        Parameters
        ----------
        type1 : AtomType or str
            If type1 is a AtomType object, then will search for its equivalent type for bond parameters.
            If type1 is a str, then will match exactly this type.
        type2 : AtomType or str
            If type2 is a AtomType object, then will search for its equivalent type for bond parameters.
            If type2 is a str, them will match exactly this type.

        Returns
        -------
        term : subclass of BondTerm
        '''
        at1: str = type1.eqt_bond if type(type1) is AtomType else type1
        at2: str = type2.eqt_bond if type(type2) is AtomType else type2
        bterm = BondTerm(at1, at2, 0)
        try:
            bterm = self.bond_terms[bterm.name]
        except:
            raise FFTermNotFoundError(f'{str(bterm)} not found in FF')

        return bterm

    def get_eqt_for_angle(self, angle):
        '''
        Get the equivalent atom types for angle parameters of the three atoms in a angle.

        Parameters
        ----------
        angle : ~mstk.topology.Angle

        Returns
        -------
        type1 : str
            The equivalent atom type for angle parameters of the first atom in this angle.
        type2 : str
            The equivalent atom type for angle parameters of the second atom in this angle.
        type3 : str
            The equivalent atom type for angle parameters of the third atom in this angle.
        '''
        try:
            at1 = self.atom_types.get(angle.atom1.type).eqt_ang_s
            at2 = self.atom_types.get(angle.atom2.type).eqt_ang_c
            at3 = self.atom_types.get(angle.atom3.type).eqt_ang_s
        except:
            raise FFTermNotFoundError(f'Atom type {angle.atom1.type} or {angle.atom2.type} '
                                      f'or {angle.atom3.type} not found in FF')
        return [(at1, at2, at3),
                (at1, at2, '*'),
                ('*', at2, at3),
                ('*', at2, '*'),
                ]

    def get_angle_term(self, type1, type2, type3):
        '''
        Get the angle term for the angle formed by three atom types.

        If not found, an Exception will be raised.

        Parameters
        ----------
        type1 : AtomType or str
            If type1 is a AtomType object, then will search for its equivalent type for angle side parameters.
            If type1 is a str, then will match exactly this type.
        type2 : AtomType or str
            If type2 is a AtomType object, then will search for its equivalent type for angle center parameters.
            If type2 is a str, them will match exactly this type.
        type3 : AtomType or str
            If type3 is a AtomType object, then will search for its equivalent type for angle side parameters.
            If type3 is a str, them will match exactly this type.

        Returns
        -------
        term : subclass of AngleTerm
        '''
        at1: str = type1.eqt_ang_s if type(type1) is AtomType else type1
        at2: str = type2.eqt_ang_c if type(type2) is AtomType else type2
        at3: str = type3.eqt_ang_s if type(type3) is AtomType else type3
        aterm = AngleTerm(at1, at2, at3, 0)
        try:
            aterm = self.angle_terms[aterm.name]
        except:
            raise FFTermNotFoundError(f'{str(aterm)} not found in FF')

        return aterm

    def get_eqt_for_dihedral(self, dihedral):
        '''
        Get the list of possible equivalent atom types for dihedral parameters of the four atoms in a dihedral.

        Because the wildcard (*) is allowed for the side atoms of a dihedral term,
        a list of possible equivalent atom types is returned.

        Parameters
        ----------
        dihedral : ~mstk.topology.Dihedral

        Returns
        -------
        types : list of tuple of str
            Each element of the list is a tuple of four str (type1, type2, type3, type4),
            which are the equivalent atom types for dihedral parameters of the four atom in this dihedral.
        '''
        try:
            at1 = self.atom_types.get(dihedral.atom1.type).eqt_dih_s
            at2 = self.atom_types.get(dihedral.atom2.type).eqt_dih_c
            at3 = self.atom_types.get(dihedral.atom3.type).eqt_dih_c
            at4 = self.atom_types.get(dihedral.atom4.type).eqt_dih_s
        except:
            raise FFTermNotFoundError(f'Atom type {dihedral.atom1.type} or {dihedral.atom2.type} '
                                      f'or {dihedral.atom3.type} or {dihedral.atom4.type} not found in FF')
        return [(at1, at2, at3, at4),
                (at1, at2, at3, '*'),
                ('*', at2, at3, at4),
                ('*', at2, at3, '*')]

    def get_dihedral_term(self, type1, type2, type3, type4):
        '''
        Get the dihedral term for the dihedral formed by four atom types.

        If not found, an Exception will be raised.

        Parameters
        ----------
        type1 : AtomType or str
            If type1 is a AtomType object, then will search for its equivalent type for dihedral side parameters.
            If type1 is a str, then will match exactly this type.
        type2 : AtomType or str
            If type2 is a AtomType object, then will search for its equivalent type for dihedral center parameters.
            If type2 is a str, them will match exactly this type.
        type3 : AtomType or str
            If type3 is a AtomType object, then will search for its equivalent type for dihedral center parameters.
            If type3 is a str, them will match exactly this type.
        type4 : AtomType or str
            If type4 is a AtomType object, then will search for its equivalent type for dihedral side parameters.
            If type4 is a str, them will match exactly this type.

        Returns
        -------
        term : subclass of DihedralTerm
        '''
        at1: str = type1.eqt_dih_s if type(type1) is AtomType else type1
        at2: str = type2.eqt_dih_c if type(type2) is AtomType else type2
        at3: str = type3.eqt_dih_c if type(type3) is AtomType else type3
        at4: str = type4.eqt_dih_s if type(type4) is AtomType else type4
        dterm = DihedralTerm(at1, at2, at3, at4)
        try:
            dterm = self.dihedral_terms[dterm.name]
        except:
            raise FFTermNotFoundError(f'{str(dterm)} not found in FF. You may use wildcard (*) for side atoms')

        return dterm

    def get_eqt_for_improper(self, improper):
        '''
        Get the list of possible equivalent atom types for improper parameters of the four atoms in a improper.

        Because the wildcard (*) is allowed for the side atoms of a improper term,
        a list of possible equivalent atom types is returned.

        Parameters
        ----------
        improper : ~mstk.topology.Improper

        Returns
        -------
        types : list of tuple of str
            Each element of the list is a tuple of four str (type1, type2, type3, type4),
            which are the equivalent atom types for improper parameters of the four atom in this improper.
        '''
        try:
            at1 = self.atom_types.get(improper.atom1.type).eqt_imp_c
            at2 = self.atom_types.get(improper.atom2.type).eqt_imp_s
            at3 = self.atom_types.get(improper.atom3.type).eqt_imp_s
            at4 = self.atom_types.get(improper.atom4.type).eqt_imp_s
        except:
            raise FFTermNotFoundError(f'Atom type {improper.atom1.type} or {improper.atom2.type} '
                                      f'or {improper.atom3.type} or {improper.atom4.type} not found in FF')
        return [(at1, at2, at3, at4),
                (at1, at2, at3, '*'),
                (at1, at2, '*', at4),
                (at1, '*', at3, at4),
                (at1, at2, '*', '*'),
                (at1, '*', at3, '*'),
                (at1, '*', '*', at4),
                (at1, '*', '*', '*')]

    def get_improper_term(self, type1, type2, type3, type4) -> ImproperTerm:
        '''
        Get the improper term for the improper formed by four atom types.

        If not found, an Exception will be raised.

        Parameters
        ----------
        type1 : AtomType or str
            If type1 is a AtomType object, then will search for its equivalent type for improper center parameters.
            If type1 is a str, then will match exactly this type.
        type2 : AtomType or str
            If type2 is a AtomType object, then will search for its equivalent type for improper side parameters.
            If type2 is a str, them will match exactly this type.
        type3 : AtomType or str
            If type3 is a AtomType object, then will search for its equivalent type for improper side parameters.
            If type3 is a str, them will match exactly this type.
        type4 : AtomType or str
            If type4 is a AtomType object, then will search for its equivalent type for improper side parameters.
            If type4 is a str, them will match exactly this type.

        Returns
        -------
        term : subclass of ImproperTerm
        '''
        at1: str = type1.eqt_imp_c if type(type1) is AtomType else type1
        at2: str = type2.eqt_imp_s if type(type2) is AtomType else type2
        at3: str = type3.eqt_imp_s if type(type3) is AtomType else type3
        at4: str = type4.eqt_imp_s if type(type4) is AtomType else type4
        iterm = ImproperTerm(at1, at2, at3, at4)
        try:
            iterm = self.improper_terms[iterm.name]
        except:
            raise FFTermNotFoundError(f'{str(iterm)} not found in FF. You may use wildcard (*) for side atoms')

        return iterm

    def get_eqt_for_qinc(self, bond):
        '''
        Get the equivalent atom types for charge increment parameters of the two atoms in a bond.

        Parameters
        ----------
        bond : ~mstk.topology.Bond

        Returns
        -------
        type1 : str
            The equivalent atom type for charge increment parameters of the first atom in this bond.
        type2 : str
            The equivalent atom type for charge increment parameters of the second atom in this bond.
        '''
        try:
            at1 = self.atom_types.get(bond.atom1.type).eqt_q_inc
            at2 = self.atom_types.get(bond.atom2.type).eqt_q_inc
        except:
            raise FFTermNotFoundError(
                f'Atom type {bond.atom1.type} or {bond.atom2.type} not found in FF')
        return at1, at2

    def get_qinc_term(self, type1, type2):
        '''
        Get the charge increment term for the bond formed by two atom types.
        A direction will also be returned, which equal to 1 if type1 in the charge increment term matches type1 in arguments, -1 otherwise.

        If not found, an Exception will be raised.

        Parameters
        ----------
        type1 : AtomType or str
            If type1 is a AtomType object, then will search for its equivalent type for charge increment parameters.
            If type1 is a str, then will match exactly this type.
        type2 : AtomType or str
            If type2 is a AtomType object, then will search for its equivalent type for charge increment parameters.
            If type2 is a str, them will match exactly this type.

        Returns
        -------
        term : subclass of BondTerm
        direction : [1, -1]
        '''
        at1: str = type1.eqt_q_inc if type(type1) is AtomType else type1
        at2: str = type2.eqt_q_inc if type(type2) is AtomType else type2
        qterm = ChargeIncrementTerm(at1, at2, 0)
        direction = 1 if qterm.type1 == at1 else -1
        try:
            qterm = self.qinc_terms[qterm.name]
        except:
            raise FFTermNotFoundError(f'{str(qterm)} not found in FF')

        return qterm, direction

    def get_all_terms(self):
        '''
        Get all the FFTerms in this force field

        Returns
        -------
        terms : list of FFTerm
        '''
        terms = []
        for d in (self.atom_types,
                  self.qinc_terms,
                  self.vdw_terms,
                  self.pairwise_vdw_terms,
                  self.bond_terms,
                  self.angle_terms,
                  self.dihedral_terms,
                  self.improper_terms,
                  self.polarizable_terms,
                  self.virtual_site_terms):
            terms.extend(d.values())
        return terms

    def assign_mass(self, top_or_mol):
        '''
        Assign masses for all atoms in a topology or molecule from the force field.

        The atom types should have been defined, and the AtomType in FF should carry mass information.
        The masses of Drude particles will be determined ONLY from the DrudeTerm,
        and the the same amount of mass will be subtracted from the parent atoms.
        That is, if there is an AtomType for Drude particles, the mass attribute of this AtomType will be simply ignored.

        Parameters
        ----------
        top_or_mol : Topology or Molecule
        '''
        _atype_not_found = set()
        _atype_no_mass = set()
        _zero_mass = set()
        top = top_or_mol
        for atom in top.atoms:
            if atom.is_drude:
                continue
            atype = self.atom_types.get(atom.type)
            if atype is None:
                _atype_not_found.add(atom.type)
                continue
            elif atype.mass == -1:
                _atype_no_mass.add(atom.type)
            elif atype.mass == 0:
                _zero_mass.add(atom.type)
            atom.mass = atype.mass

        if _atype_not_found != set():
            logger.error('%i atom types not found: %s' % (
                len(_atype_not_found), ' '.join(_atype_not_found)))
            raise Exception(f'Assign mass for {str(top)} failed')
        if _atype_no_mass != set():
            logger.error('%i atom types in FF do not carry mass information: %s' % (
                len(_atype_no_mass), ' '.join(_atype_no_mass)))
            raise Exception(f'Assign mass {str(top)}  failed')
        if _zero_mass != set():
            logger.warning(f'%i atoms in {str(top)} carry zero mass: %s' % (
                len(_zero_mass), ' '.join(_zero_mass)))

        for parent, drude in top.get_drude_pairs():
            atype = self.atom_types[parent.type]
            pterm = self.polarizable_terms.get(atype.eqt_polar)
            if pterm is None:
                raise Exception(f'Polarizable term for {atype} not found in FF')
            if type(pterm) != DrudeTerm:
                raise Exception('Polarizable terms other than DrudeTerm haven\'t been implemented')
            drude.mass = pterm.mass
            parent.mass -= pterm.mass

    def assign_charge(self, top_or_mol, transfer_qinc_terms=False):
        '''
        Assign charges for all atoms in a topology or molecule from the force field.

        The atom types should have been defined.
        The charge of corresponding AtomType in FF will be assigned to the atoms first.
        Then if there are ChargeIncrementTerm, the charge will be transferred between bonded atoms.
        The charge of Drude particles will be determined ONLY from the DrudePolarizableTerm,
        and the same amount of charge will be subtracted from the parent atoms.
        That is, if there is an AtomType for Drude particles, the charge attribute of this AtomType will be simply ignored.

        If charge increment parameters required by the topology are not found in the force field,
        `mstk` can try to transfer parameters from more general terms in the force field.
        e.g. if charge increment term between 'c_4o2' and 'n_3' is not found, the term between 'c_4' and 'n_3' will be used.
        This mechanism is initially designed for TEAM force field.
        It is disabled by default, and can be enabled by setting argument `transfer_qinc_terms` to True.

        Parameters
        ----------
        top_or_mol : Topology or Molecule
        transfer_qinc_terms : bool, optional
        '''
        from .dff_utils import dff_fuzzy_match

        _q_zero = []
        top = top_or_mol
        for atom in top.atoms:
            if atom.is_drude:
                continue
            try:
                charge = self.atom_types[atom.type].charge
            except:
                raise Exception(f'Atom type {atom.type} not found in FF')
            atom.charge = charge
            if charge == 0:
                _q_zero.append(atom)

        if len(self.qinc_terms) == 0 and len(_q_zero) > 0:
            msg = '%i atoms in %s have zero charge. But charge increment terms not found in FF: %s' % (
                len(_q_zero), top, ' '.join([a.name for a in _q_zero[:10]]))
            if len(_q_zero) > 10:
                msg += ' and more ...'
            logger.warning(msg)

        if len(self.qinc_terms) > 0:
            _qterm_not_found = set()
            _qterm_transferred = set()
            for bond in filter(lambda x: not x.is_drude, top.bonds):
                _found = False
                ats = self.get_eqt_for_qinc(bond)
                if ats[0] == ats[1]:
                    continue
                _term = ChargeIncrementTerm(*ats, 0)
                try:
                    qterm, direction = self.get_qinc_term(*ats)
                    _found = True
                except FFTermNotFoundError:
                    if transfer_qinc_terms:
                        qterm, score = dff_fuzzy_match(_term, self)
                        if qterm is not None:
                            direction = 1 if _term.type1 == ats[0] else -1
                            _found = True
                            _qterm_transferred.add((_term.name, qterm.name, score))

                if _found:
                    increment = qterm.value * direction
                    bond.atom1.charge += increment
                    bond.atom2.charge -= increment
                else:
                    _qterm_not_found.add(_term.name)

            if len(_qterm_transferred) > 0:
                logger.warning('%i charge increment terms not found in FF. Transferred from similar terms with score\n'
                               '        %s' % (len(_qterm_transferred),
                                               '\n        '.join(['%-16s %-16s %i' % x for x in _qterm_transferred])))
            if len(_qterm_not_found) > 0:
                logger.warning('%i charge increment terms not found in FF. Make sure FF is correct\n'
                               '        %s' % (len(_qterm_not_found),
                                               '\n        '.join(_qterm_not_found)))

        for parent, drude in top.get_drude_pairs():
            atype = self.atom_types[parent.type]
            pterm = self.polarizable_terms.get(atype.eqt_polar)
            if pterm is None:
                raise Exception(f'Polarizable term for {atype} not found in FF')
            if type(pterm) != DrudeTerm:
                raise Exception('Polarizable terms other than DrudeTerm haven\'t been implemented')
            n_H = len([atom for atom in parent.bond_partners if atom.symbol == 'H'])
            alpha = pterm.alpha + n_H * pterm.merge_alpha_H
            drude.charge = -pterm.get_charge(alpha)
            parent.charge += pterm.get_charge(alpha)
            # update alpha and thole for Drude parent particle
            parent.alpha = alpha
            parent.thole = pterm.thole



import os
import shutil
import subprocess
from subprocess import Popen, PIPE

from ..errors import GmxError
from ..utils import random_string


class GMX:
    '''
    Wrapper for GROAMCS for pre-processing, running and post-processing MD simulations.

    GROMACS version 2016 is recommended and has been tested thoroughly.
    Though this wrapper support both version 2016, 2018 and 2019.

    It's better to compile GROMACS in two binaries.
    One is the normal `gmx` binary without any acceleration.
    This is used for pre- and post-processing, which normally run on the front node of a HPC.
    The other one is the `mdrun` binary with full accelerations through MPI, OpenMP, GPU, etc...
    This will be submitted to :class:`~mstk.scheduler.Scheduler` to run on the compute node of a HPC.
    Refer to GROMACS documentation for the compiling options.

    If gmx_bin and gmx_mdrun are provided separately (recommended), you should make sure that they have the same version.

    Parameters
    ----------
    gmx_bin : str
        Path of the serial version gmx binary executable.
    gmx_mdrun : str, optional
        Path of the parallel/gpu version mdrun binary executable.
        If not provided, `gmx_bin` will be used also for mdrun.
    version : str, optional
        The version of GROMACS, e.g. 2016.6, 2019.6.
        If not specified, then gmx_bin will be called to check the version.
        In this case, make sure gmx_bin is available on this machine.

    Attributes
    ----------
    version : str
    majorversion : str
    '''

    _TEMPLATE_DIR = os.path.abspath(os.path.dirname(__file__) + os.sep + '../template/gmx/')

    def __init__(self, gmx_bin, gmx_mdrun=None, version=None):
        self.GMX_BIN = gmx_bin
        self.GMX_MDRUN = gmx_mdrun or gmx_bin + ' mdrun'
        if version is None:
            self._check_version()
        else:
            self.version = version
            self.majorversion = self.version.split('.')[0]

        # TODO temporary hack for dielectric constant in mdp
        self._DIELECTRIC = 1.0
        # TODO temporary hack for LJ96 function
        self._LJ96 = False

    def _check_version(self):
        cmd = '%s --version' % self.GMX_BIN
        try:
            sp = Popen(cmd.split(), stdout=PIPE, stderr=PIPE)
        except:
            raise GmxError('gmx not valid')
        stdout = sp.communicate()[0]
        for line in stdout.decode().splitlines():
            if line.startswith('GROMACS version'):
                break
        else:
            raise GmxError('gmx not valid')

        self.version = line.strip().split()[-1]
        if not (self.version.startswith('2016') or self.version.startswith(
                '2018') or self.version.startswith('2019')):
            raise GmxError('Supported GROMACS versions: 2016.x, 2018.x, 2019.x')

        self.majorversion = self.version.split('.')[0]

    def grompp(self, mdp='grompp.mdp', gro='conf.gro', top='topol.top', tpr_out='md.tpr',
               cpt=None, maxwarn=3, silent=False, get_cmd=False):
        '''
        Run `gmx grompp` directly or generate the command for `gmx grompp`.

        Refer to GROMACS documentation for details.

        Parameters
        ----------
        mdp : str
        gro : str
        top : str
        tpr_out : str
        cpt : str
        maxwarn : int
        silent : bool
            If set to True and `get_cmd` set to False, then `gmx grompp` will be executed silently, without output on screen.
        get_cmd : bool
            If set to True, the command for running `gmx grompp` will be returned, which can be feed into :class:`~mstk.scheduler.Scheduler`.
            If set to False, `gmx grompp` will be executed.

        Returns
        -------
        command : str or None
            If get_cmd set to True, will return the command for running `gmx grompp`.
            If get_cmd set to False, will return None.

        '''
        cmd = '%s -quiet -nobackup grompp -f %s -c %s -p %s -o %s -maxwarn %i' % (
            self.GMX_BIN, mdp, gro, top, tpr_out, maxwarn)
        if cpt is not None:
            cmd = '%s -t %s' % (cmd, cpt)
        if get_cmd:
            return cmd
        else:
            (stdout, stderr) = (PIPE, PIPE) if silent else (None, None)
            sp = Popen(cmd.split(), stdout=stdout, stderr=stderr)
            sp.communicate()

    def mdrun(self, name='md', nprocs=1, n_omp=None, rerun=None, extend=False, silent=False,
              get_cmd=False):
        '''
        Run `mdrun` directly or generate the command for `mdrun`.

        `mdrun` will use both MPI and OpenMP for hybrid parallelization.

        Refer to GROMACS documentation for details.

        Parameters
        ----------
        name : str
            Argument for `mdrun -deffnm`
        nprocs : int
            Number of CPU cores to use
        n_omp : int, optional
            Number of OpenMP threads to use.
            If not set, will be determined automatically from nprocs.
        rerun : str, optional
            If an XTC or TRR file privided, will perform `mdrun -rerun` on the trajectory.
        extend : bool
            If set to True, will perform `mdrun -cpi` on the checkpoint.
            A checkpoint named `name.cpt` should be avaiable in the current directory.
        silent : bool
            If set to True and `get_cmd` set to False, then `mdrun` will be executed silently, without output on screen.
        get_cmd : bool
            If set to True, the command for running `mdrun` will be returned, which can be feed into :class:`~mstk.scheduler.Scheduler`.
            If set to False, `mdrun` will be executed.

        Returns
        -------
        command : str or None
            If get_cmd set to True, will return the command for running `mdrun`.
            If get_cmd set to False, will return None.

        '''
        # TODO temporary hack for LJ96 function
        if self._LJ96:
            n_omp = 1
        ###

        if n_omp is None:  # n_omp is None means auto
            for i in [6, 4, 2]:  # available OpenMP threads: 6, 4, 2
                if nprocs % i == 0:
                    n_omp = i
                    break
            else:
                n_omp = nprocs
        n_mpi = nprocs // n_omp

        # TODO temporary hack for LJ96 function
        if self._LJ96:
            if name.find('hvap') != -1:
                n_mpi = 1
        ###

        cmd = '%s -quiet -nobackup -ntomp %i -deffnm %s' % (self.GMX_MDRUN, n_omp, name)
        # always use mpirun even if only one process
        cmd = 'mpirun -np %i ' % n_mpi + cmd

        if rerun is not None:
            cmd = '%s -rerun %s' % (cmd, rerun)

        if extend:
            cmd = '%s -cpi %s' % (cmd, name + '.cpt')

        if get_cmd:
            return cmd
        else:
            (stdout, stderr) = (PIPE, PIPE) if silent else (None, None)
            sp = Popen(cmd.split(), stdout=stdout, stderr=stderr)
            sp.communicate()

    def prepare_mdp_from_template(self, template, mdp_out='grompp.mdp', T=298, P=1, nsteps=10000,
                                  dt=0.001, TANNEAL=800,
                                  nstenergy=100, nstxout=0, nstvout=0, nstxtcout=10000,
                                  xtcgrps='System',
                                  restart=False, tcoupl='langevin', pcoupl='parrinello-rahman',
                                  constraints='h-bonds', ppm=0, dielectric=None):
        '''
        Generate MDP control script from template library for running GROMACS simulation.

        Refer to GROMACS documentation for the meaning of parameteres.

        The coupling time constant are determined automatically from `dt`, `tcoupl` and `pcoupl`.
        The neighbour list update frequency is determined automatically from `dt`.

        Parameters
        ----------
        template : str
            File name of the template to use.
            Take a look of the template directory to get available choices.
        mdp_out : str
        T : float
        P : float
        nsteps : int
        dt : float
        TANNEAL : float, optional
            The temperature for annealing.
            This argument should be used together with template `t_nvt_anneal.mdp`
            It is ignored for other templates.
        nstenergy : int
        nstxout : int
        nstvout : int
        nstxtcout : int
        xtcgrps : int
        restart : bool
            If set to True, grompp will not generated velocity.
        tcoupl : ['langevin', 'v-rescale', 'nose-hoover']
        pcoupl : ['parrinello-rahman', 'berendsen']
        constraints : ['h-bonds', 'none', 'all-bonds']
        ppm : float, optional
            The cosine acceleration strength for periodic perturbation simulation.
            This argument should be used together with template `t_npt_ppm.mdp`
            It is ignored for other templates.
        dielectric : float, optional

        '''
        template = os.path.join(GMX._TEMPLATE_DIR, template)
        if not os.path.exists(template):
            raise GmxError('mdp template not found')

        if tcoupl.lower() == 'langevin':
            integrator = 'sd'
            tcoupl = 'no'
            tau_t = str(0.001 / dt)  # inverse friction coefficient
        elif tcoupl.lower() == 'nose-hoover':
            integrator = 'md'
            tcoupl = 'nose-hoover'
            tau_t = '0.5'
        elif tcoupl.lower() == 'v-rescale':
            integrator = 'md'
            tcoupl = 'v-rescale'
            tau_t = '0.1'
        else:
            raise Exception('Invalid tcoupl, should be one of langvein, nose-hoover, v-rescale')

        if pcoupl.lower() == 'berendsen':
            tau_p = '1'
        elif pcoupl.lower() == 'parrinello-rahman':
            tau_p = '5'
        else:
            raise Exception('Invalid pcoupl, should be one of berendsen, parrinello-rahman')

        if restart:
            genvel = 'no'
            continuation = 'yes'
        else:
            genvel = 'yes'
            continuation = 'no'

        nstlist = max(1, int(0.01 / dt))

        if dielectric is None:
            dielectric = self._DIELECTRIC

        with open(template) as f_t:
            contents = f_t.read()
        contents = contents.replace('%T%', str(T)).replace('%P%', str(P)).replace('%nsteps%',
                                                                                  str(int(nsteps))) \
            .replace('%dt%', str(dt)).replace('%nstenergy%', str(nstenergy)) \
            .replace('%nstxout%', str(nstxout)).replace('%nstvout%', str(nstvout)) \
            .replace('%nstxtcout%', str(nstxtcout)).replace('%xtcgrps%', str(xtcgrps)) \
            .replace('%genvel%', genvel).replace('%continuation%', continuation) \
            .replace('%integrator%', integrator).replace('%tcoupl%', tcoupl).replace('%tau-t%',
                                                                                     tau_t) \
            .replace('%pcoupl%', pcoupl).replace('%tau-p%', tau_p) \
            .replace('%constraints%', constraints).replace('%TANNEAL%', str(TANNEAL)).replace(
            '%ppm%', str(ppm)) \
            .replace('%nstlist%', str(nstlist)).replace('%dielectric%', str(dielectric))

        with open(mdp_out, 'w') as f_mdp:
            f_mdp.write(contents)

    def energy(self, edr, properties, begin=0, end=None, fluct_props=False, get_cmd=False):
        '''
        Execute `gmx energy` program on a `edr` file.

        Refer to the GROMACS documentation for details.

        Parameters
        ----------
        edr : str
        properties : list of str
            The properties to be extracted.
            It corresponds to the input when calling gmx_energy manually.
            But instead of input one by one joined by Enter, the properties are provided in a list here.
        begin : float
        end : float, optional
        fluct_props : bool
        get_cmd : bool

        Returns
        -------
        out : str
            If set to True, the command for running `gmx energy` will be returned.
            If set to False, `gmx energy` will be executed silently and the output will be returned directly without cleanup.
        '''
        cmd = '%s -quiet -nobackup energy -f %s -b %s' % (self.GMX_BIN, edr, str(begin))
        if end is not None:
            cmd += ' -e %s' % (str(end))
        if fluct_props:
            cmd += ' -fluct_props'
        if get_cmd:
            property_str = '\\n'.join(properties)
            cmd = 'echo "%s" | %s' % (property_str, cmd)
            return cmd
        else:
            sp = Popen(cmd.split(), stdout=PIPE, stdin=PIPE, stderr=PIPE)
            property_str = '\n'.join(properties)
            out, err = sp.communicate(input=property_str.encode())
            return out.decode()

    def get_fluct_props(self, edr, begin=0, end=None):
        '''
        Get thermal expansion and compressibility from a `edr` file using fluctuation of enthalpy, volume.

        Parameters
        ----------
        edr : str
        begin : float
        end : float, optional

        Returns
        -------
        expansion : float
        compressibility : float
        '''
        sp_out = self.energy(edr, properties=['temp', 'vol', 'enthalpy'], begin=begin, end=end,
                             fluct_props=True)

        expansion = None
        compressibility = None
        for line in sp_out.splitlines():
            if line.startswith('Coefficient of Thermal Expansion Alpha_P'):
                expansion = float(line.split()[-2])
            elif line.startswith('Isothermal Compressibility Kappa'):
                compressibility = float(line.split()[-2]) * 1e5
        return expansion, compressibility

    def get_properties_stderr(self, edr, properties, begin=0, end=None):
        '''
        Extract properties from a `edr` file by calling `gmx energy` program.
        The difference between this method and :func:`energy` is that,
        :func:`energy` will return the raw output string without any cleanup,
        while this method will extract the values and standard errors from the output and return only the results.

        Parameters
        ----------
        edr : str
        properties : list of str
        begin : float
        end : float, optional

        Returns
        -------
        results : list of tuple of float
            Each element of this list is a tuple of two floats, which are average value and standard error of each property.
        '''
        sp_out = self.energy(edr, properties=properties, begin=begin, end=end)

        lines = sp_out.splitlines()
        results = []
        for prop in properties:
            for line in lines:
                if line.lower().startswith(prop.lower()):
                    results.append((float(line.split()[1]), float(line.split()[2])))
                    break
            else:
                raise GmxError('Invalid property')
        return results

    def get_property_stderr(self, edr, prop, begin=0, end=None):
        '''
        Extract the average value and standard error of a property from a `edr` file
        by calling `gmx energy` program.

        Parameters
        ----------
        edr : str
        prop: str
        begin : float
        end : float, optional

        Returns
        -------
        average : float
        stderr : float
        '''
        return self.get_properties_stderr(edr, [prop], begin, end)[0]

    def get_box(self, edr, begin=0):
        '''
        Extract the average box size from a `edr` file by calling `gmx energy` program.

        Only rectangular box is supported.

        Parameters
        ----------
        edr : str
        begin : float

        Returns
        -------
        lx : float
        ly : float
        lz : float
        '''
        sp = subprocess.Popen([self.GMX_BIN, 'energy', '-f', edr, '-b', str(begin)],
                              stdout=PIPE, stdin=PIPE, stderr=PIPE)
        sp_out = sp.communicate(input='Box-'.encode())[0]

        box = [0, 0, 0]
        for line in sp_out.decode().splitlines():
            if line.startswith('Box-X'):
                box[0] = float(line.split()[1])
            if line.startswith('Box-Y'):
                box[1] = float(line.split()[1])
            if line.startswith('Box-Z'):
                box[2] = float(line.split()[1])
        return box

    def density(self, trj, tpr, xvg='density.xvg', group='System', begin=0, end=None,
                center=False, silent=False, get_cmd=False):
        '''
        Execute `gmx density` to calculate the density profile in `z` direction on a trajectory.

        Refer to GROMACS documentation for details.

        Parameters
        ----------
        trj : str
        tpr : str
        xvg : str
        group : str
        begin : float
        end : float, optional
        center : bool
        silent : bool
            If set to True and `get_cmd` set to False, then `gmx density` will be executed silently, without output on screen.
        get_cmd : bool
            If set to True, the command for running `gmx density` will be returned, which can be feed into :class:`~mstk.scheduler.Scheduler`.
            If set to False, `gmx density` will be executed.

        Returns
        -------
        command : str or None
            If get_cmd set to True, will return the command for running `gmx_density`.
            If get_cmd set to False, will return None.
        '''
        cmd = '%s -quiet -nobackup density -f %s -s %s -b %f -o %s' % (
            self.GMX_BIN, trj, tpr, begin, xvg)
        if end is not None:
            cmd += ' -e %f' % end
        inp = group
        if center:
            cmd += ' -center'
            inp = '%s\n%s' % (group, group)

        if get_cmd:
            cmd = 'echo "%s" | %s' % (inp, cmd)
            return cmd
        else:
            (stdout, stderr) = (PIPE, PIPE) if silent else (None, None)
            sp = Popen(cmd.split(), stdin=PIPE, stdout=stdout, stderr=stderr)
            sp.communicate(input=inp.encode())

    @staticmethod
    def scale_box(gro, gro_out, new_box):
        '''
        Scale GRO box to desired size.

        The coordinate of all atoms are scaled, while the velocities are not modified.
        Only rectangular box is supported.

        Parameters
        ----------
        gro : str
        gro_out : str
        new_box : list of float
            The lengths of the desired rectangular box.
        '''
        NAtoms = 0
        nresidue = []
        residue = []
        element = []
        natom = []
        xyz = []
        vxyz = []
        box = []
        if not os.path.exists(gro):
            raise GmxError('gro not found')
        with open(gro) as f_gro:
            lines = f_gro.read().splitlines()

        for i, line in enumerate(lines):
            n = i + 1
            if n == 1:
                title = line
            if n == 2:
                NAtoms = int(line.strip())
                continue
            if n > 2 and n <= 2 + NAtoms:
                nresidue.append(int(line[:5]))
                residue.append(line[5:10])
                element.append(line[10:15])
                natom.append(int(line[15:20]))
                x = float(line[20:28])
                y = float(line[28:36])
                z = float(line[36:44])
                vx = float(line[44:52])
                vy = float(line[52:60])
                vz = float(line[60:68])
                xyz.append([x, y, z])
                vxyz.append([vx, vy, vz])
            if n == 3 + NAtoms:
                box = [float(word) for word in line.strip().split()[:3]]
                break

        scale = [new_box[i] / box[i] for i in range(3)]
        xyz = [[i[0] * scale[0], i[1] * scale[1], i[2] * scale[2]] for i in xyz]

        with open(gro_out, 'w') as f_out:
            f_out.write('Scaled : %s\n%i\n' % (title, NAtoms))
            for i in range(NAtoms):
                f_out.write('%5i%5s%5s%5i%8.3f%8.3f%8.3f%8.4f%8.4f%8.4f\n'
                            % (nresidue[i], residue[i], element[i], natom[i],
                               xyz[i][0], xyz[i][1], xyz[i][2], vxyz[i][0], vxyz[i][1], vxyz[i][2]))
            f_out.write('%f %f %f\n' % (new_box[0], new_box[1], new_box[2]))

    @staticmethod
    def generate_top(itp, molecules, numbers):
        '''
        Generate a `topol` file that includes a existent `itp` file.

        Parameters
        ----------
        itp : str
        molecules : list of str
            Name of molecules in the system
        numbers : list of int
            Number of these molecules in the system
        '''
        shutil.copy(itp, 'topol.top')
        with open('topol.top', 'a') as f:
            f.write('\n[system]\n[molecules]\n')
            for i, molecule in enumerate(molecules):
                f.write('%s %i\n' % (molecule, numbers[i]))

    @staticmethod
    def get_top_mol_numbers(top):
        '''
        Get the name and number of molecules in a topol file from the `[ molecules ]` section.

        Parameters
        ----------
        top : str

        Returns
        -------
        mol_numbers : list of tuple
            Each element is a tuple containing a str and a int, which are the name and number of each molecule specie.

        '''
        with open(top) as f:
            lines = f.read().splitlines()

        newlines = []
        mols = []
        START = False
        for line in lines:
            if line.find('[') != -1 and line.find('molecules') != -1:  # [ molecules ]
                START = True
                newlines.append(line)
                continue
            if not START:
                newlines.append(line)
            if START and not line.strip() == '' and not line.startswith(';'):
                words = line.strip().split()
                mols.append((words[0], int(words[1])))
        return mols

    @staticmethod
    def modify_top_mol_numbers(top, numbers):
        '''
        Modify the number of molecules in a `topol` file at the `[ molecules ]` section.

        The length of numbers must be equal to the original length of numbers in the `topol` file.
        Otherwise, an Exception will be raised.

        Parameters
        ----------
        top : str
        numbers : list of int
        '''
        with open(top) as f:
            lines = f.read().splitlines()

        newlines = []
        mols = []
        START = False
        for line in lines:
            if line.find('[') != -1 and line.find('molecules') != -1:  # [ molecules ]
                START = True
                newlines.append(line)
                continue
            if not START:
                newlines.append(line)
            if START and not line.strip() == '' and not line.startswith(';'):
                words = line.strip().split()
                mols.append((words[0], int(words[1])))
        if len(mols) != len(numbers):
            raise GmxError('Type of molecules in top not consistent')

        for i, (mol_name, _) in enumerate(mols):
            newlines.append('%s\t%i' % (mol_name, numbers[i]))

        with open(top, 'w')  as f:
            f.write('\n'.join(newlines))

    def replicate_gro(self, gro, top, nbox, silent=True):
        '''
        Enlarge a GRO box by replicate it with `gmx genconf`.

        The numbers of molecules in the `topol` file will be updated accordingly.

        Parameters
        ----------
        gro : str
        top : str
        nbox : list of int
            Three element denotes the number of replicates in x, y  and z directions.
        silent : bool
        '''
        from functools import reduce
        import operator

        gro_tmp = random_string(8) + '.gro'
        cmd = '%s -quiet genconf -f %s -o %s -nbox %i %i %i' % (
            self.GMX_BIN, gro, gro_tmp, nbox[0], nbox[1], nbox[2])
        (stdout, stderr) = (PIPE, PIPE) if silent else (None, None)
        sp = Popen(cmd.split(), stdout=stdout, stderr=stderr)
        sp.communicate()
        shutil.move(gro_tmp, gro)

        with open(top) as f:
            content = f.read()

        LASTLINE = '\n'
        LAST = False
        for line in content.splitlines():
            if line.find('molecules') > -1 and line.find('[') > -1:
                LAST = True
                continue
            if LAST and not line.strip() == '' and not line.startswith(';'):
                LASTLINE += line + '\n'
        LASTLINE *= (reduce(operator.mul, nbox, 1) - 1)

        with open(top, 'a') as f:
            f.write(LASTLINE)

    def pdb2gro(self, pdb, gro_out, box, silent=False):
        '''
        Convert a PDB file to GRO file by calling `gmx editconf`.

        Parameters
        ----------
        pdb : str
        gro_out : str
        box : list of float
            The size of rectangular box.
        silent : bool
        '''
        if len(box) != 3:
            raise GmxError('Invalid box')

        (stdout, stderr) = (PIPE, PIPE) if silent else (None, None)
        sp = Popen(
            [self.GMX_BIN, 'editconf', '-f', pdb, '-o', gro_out, '-box', str(box[0]), str(box[1]),
             str(box[2])],
            stdin=PIPE, stdout=stdout, stderr=stderr)
        sp.communicate()

    def velacc(self, trr, tpr=None, group=None, begin=0, xvg_out='velacc', silent=False):
        '''
        Calculte the velocity auto correlation of atoms in a trajectory by calling `gmx velacc`.

        Refer to GROMACS documentation for details.

        Parameters
        ----------
        trr : str
        tpr : str
        group : str
        begin : float
        xvg_out : str
        silent : bool
        '''
        if tpr is None:
            tpr = trr
        if group is None:
            raise GmxError('No group specifed')

        (stdout, stderr) = (PIPE, PIPE) if silent else (None, None)
        sp = Popen(
            [self.GMX_BIN, 'velacc', '-f', trr, '-s', tpr, '-o', xvg_out, '-b', str(begin), '-mol',
             '-nonormalize'],
            stdin=PIPE, stdout=stdout, stderr=stderr)
        sp.communicate(input=str(group).encode())

    def diffusion(self, xtc, tpr, group='System', mol=False, begin=0, end=None, xvg_out='msd.xvg',
                  beginfit=-1, endfit=-1):
        '''
        Calculate the diffusion coefficient from a trajectory file by calling `gmx msd`.

        Refer to GROMACS documentation for details.

        Parameters
        ----------
        xtc : str
        tpr : str
        group : str
        mol : bool
        begin : float
        end : float, optional
        xvg_out : str
        beginfit : float, optional
        endfit : str, optional

        Returns
        -------
        diffusion : float
            Diffusion coefficient in unit of cm^2/s.
        stderr : float
            Standard error.
        '''
        cmd = '%s -quiet -nobackup msd -f %s -s %s -o %s -b %s -beginfit %s -endfit %s' % (
            self.GMX_BIN, xtc, tpr, xvg_out, str(begin), str(beginfit), str(endfit))
        if end is not None:
            cmd += ' -e %s' % str(end)
        if mol:
            # calculate the MSD of COM of molecules
            cmd += ' -mol'

        sp = Popen(cmd.split(), stdin=PIPE, stdout=PIPE, stderr=PIPE)
        out, err = sp.communicate(input=str(group).encode())

        for line in out.decode().splitlines():
            if line.startswith('D['):
                words = line.strip().split(']')[-1].strip().split()
                diffusion = float(words[0])
                stderr = float(words[2][:-1])
                unit = float(words[3])
                diffusion *= unit  # cm^2/s
                stderr *= unit  # cm^2/s
                return diffusion, stderr

        raise GmxError('Error running gmx msd')

    def traj_com(self, xtc, tpr, trj_out='com.xtc', begin=0, end=0, silent=False):
        '''
        Generate a trajectory of the centers of mass of molecules in a trajectory.

        It calls :func:`select_com` to create a index file of the centers of mass of all molecules.
        Then `gmx traj` is called to calculate the positions of each center of mass and output the trajectory.

        Refer to the GROMACS documentation for details.

        Parameters
        ----------
        xtc : str
        tpr : str
        trj_out : str
        begin : float
        end : float
        silent : bool
        '''
        ndx = 'com.ndx'
        self.select_com(tpr, 'all', ndx_out=ndx)

        (stdout, stderr) = (PIPE, PIPE) if silent else (None, None)
        cmd = '%s -quiet -nobackup traj -s %s -f %s -oxt %s -com -mol -n %s -b %f -e %f' % (
            self.GMX_BIN, tpr, xtc, trj_out, ndx, begin, end)
        sp = Popen(cmd.split(), stdout=stdout, stderr=stderr)
        sp.communicate()
        return trj_out, ndx

    def select_com(self, tpr, resname='all', ndx_out='com.ndx'):
        '''
        Create an index file of the centers of mass of each molecule by calling `gmx select`.

        Refer to GROAMCS documentation for details.

        Parameters
        ----------
        tpr : str
        resname : str
        ndx_out : str
        '''
        cmd = '%s -quiet -nobackup select -s %s -on %s' % (self.GMX_BIN, tpr, ndx_out)
        sp = Popen(cmd.split(), stdin=PIPE, stdout=PIPE, stderr=PIPE)
        if resname == 'all':
            select_com_str = 'res_com of all'
        else:
            select_com_str = 'res_com of resname %s' % resname
        sp.communicate(input=select_com_str.encode())

    @staticmethod
    def generate_top_for_hvap(top, top_out):
        '''
        Generate a new `topol` file for calculating the heat of vaporization from existent `topol` file for bulk simulation.

        Essentially, this will add a `[ exclusion ]` section in the `topol` file for each molecule.
        All intra-molecular interactions will be excluded.
        With this modified `topol` file, `gmx mdrun -rerun` on the bulk trajectory will calculate
        only the inter-molecular energies, which is approximately the cohesive energy.

        Parameters
        ----------
        top : str
        top_out : str
        '''

        with open(top) as f:
            lines = f.read().splitlines()
        lines = [l for l in lines if not (l.startswith(';') or l == '')]

        f_out = open(top_out, 'w')

        line_number_molecule = []
        line_number_atom = []
        line_number_system = None
        for n, line in enumerate(lines):
            if line.find('[') != -1 and line.find('moleculetype') != -1:
                line_number_molecule.append(n)
            if line.find('[') != -1 and line.find('atoms') != -1:
                line_number_atom.append(n)
            if line.find('[') != -1 and line.find('system') != -1:
                line_number_system = n

        n_molecules = len(line_number_molecule)

        for n in range(line_number_molecule[0]):
            f_out.write(lines[n] + '\n')

        for i in range(n_molecules):
            for n in range(line_number_molecule[i], line_number_atom[i]):
                f_out.write(lines[n] + '\n')
            line_number_next_section = line_number_molecule[i + 1] if i < n_molecules - 1 else line_number_system
            if line_number_next_section is None:
                line_number_next_section = len(lines)
            n_atoms = 0
            f_out.write('[ atoms ]\n')
            for n in range(line_number_atom[i] + 1, line_number_next_section):
                line = lines[n]
                if line.find('[') != -1 or line.startswith('#'):
                    f_out.write('[ bonds ]\n[ exclusions ]\n')
                    for i in range(1, n_atoms):
                        exclusions = range(i, n_atoms + 1)
                        f_out.write(' '.join(list(map(str, exclusions))) + '\n')
                    break
                f_out.write(line + '\n')
                n_atoms += 1

        if line_number_system is not None:
            for n in range(line_number_system, len(lines)):
                f_out.write(lines[n] + '\n')

    def extend_tpr(self, tpr, extend, silent=False):
        '''
        Modify the `nsteps` record in a `tpr` file by calling `gmx convert-tpr` for extending simulation.

        Refer to GROMACS documentation for details.

        Parameters
        ----------
        tpr : str
        extend : float
        silent : bool
        '''
        cmd = '%s -quiet convert-tpr -s %s -o %s -extend %s' % (self.GMX_BIN, tpr, tpr, str(extend))
        (stdout, stderr) = (PIPE, PIPE) if silent else (None, None)
        sp = Popen(cmd.split(), stdout=stdout, stderr=stderr)
        sp.communicate()

    def generate_gpu_multidir_cmds(self, dirs, commands, n_parallel, n_gpu=0, n_omp=None,
                                   n_procs=None):
        '''
        Generate commands for performing `multidir` simulation with GROMACS.

        multidir simulation enables several similar simulations being run at the same time in different directories.
        This is a powerful feature allowing for much better utilization of GPU resources.

        Parameters
        ----------
        dirs : list of str
        commands : list of str
        n_parallel : int
        n_gpu : int, optional
        n_omp : int, optional
            Set n_omp in most case.
            When n_procs is set, n_omp will have no effect.
        n_procs : int, optional

        Returns
        -------
        commands_list : list of list of str

        '''
        import math, re
        def replace_gpu_multidir_cmd(dirs: [str], cmd: str) -> str:
            n_multi = len(dirs)
            if cmd.startswith('export '):
                pass

            elif cmd.find('gmx') != -1 and cmd.find(' mdrun ') != -1:
                n_mpi = n_multi
                n_thread = n_omp

                if n_procs is not None:
                    if cmd.find('mpirun ') != -1 and cmd.find('mpirun -np 1 ') == -1:
                        # optimize n_mpi and n_thread
                        # GPU tasks prefer more n_mpi
                        # ensure n_mpi equals 2*n
                        # do not optimize for mpirun -np 1. This happens for rerun hvap
                        optimal_omp = [6, 4, 2] if n_gpu > 0 else [8, 6, 4, 2]
                        for i in optimal_omp:
                            if n_procs % (n_multi * i * 2) == 0:
                                n_thread = i
                                n_mpi = n_procs // n_thread
                                break

                    else:
                        # set n_mpi equal to n_multi
                        n_thread = n_procs // n_mpi

                cmd = re.sub(r'mpirun\s+-np\s+[0-9]+', '', cmd)  # remove mpirun -np xx
                cmd = re.sub(r'-ntomp\s+[0-9]+', '', cmd)  # remove -ntomp xx

                cmd = 'mpirun -np %i %s' % (n_mpi, cmd)  # add mpirun -np xx
                cmd += ' -multidir ' + ' '.join(dirs)  # add -multidir xx xx xx

                ### meaning of -gpu_id is changed in GROMACS 2018. Disable gpu assignment
                if n_gpu > 0 and self.majorversion == '2016':
                    cmd += ' -gpu_id ' + ''.join(map(str, range(n_gpu))) * (n_mpi // n_gpu) \
                           + ''.join(map(str, range(n_mpi % n_gpu)))  # add -gpu_id 01230123012
                if n_thread is not None:
                    cmd += ' -ntomp %i' % n_thread

            else:
                cmd = 'for i in %s; do cd $i; %s; done' % (
                    ' '.join(dirs), cmd)  # do it in every directory
            return cmd

        commands_list: [[str]] = []
        n_group: int = math.ceil(len(dirs) / n_parallel)
        for n in range(n_group):
            multidir = dirs[n * n_parallel:(n + 1) * n_parallel]
            commands_multidir: [str] = []
            for j, dirname in enumerate(multidir):
                commands_multidir.append(
                    'dir%i=%s' % (j, dirname))  # replace full dir path with $dir1, $dir2 ...
            for cmd in commands:
                commands_multidir.append(
                    replace_gpu_multidir_cmd(['$dir%i' % j for j in range(len(multidir))], cmd))
            commands_list.append(commands_multidir)
        return commands_list

    @staticmethod
    def modify_lj96(itp_files, top_files, mdp_files, xvg_files):
        '''
        Convert the current simulation files for LJ-12-6 vdW form into LJ-9-6 form.

        This is a trick to simplify the exporting of LJ-9-6 vdW form in GROMACS.

        Parameters
        ----------
        itp_files : None or list of str
            The `itp` files to be modified.
            If not set, all the `itp` files in the current directory will be modified.
        top_files : None or list of str
            The `top` files to be modified.
            If not set, all the `top` files in the current directory will be modified.
        mdp_files : None or list of str
            The `mdp` files to be modified.
            If not set, all the `mdp` files in the current directory will be modified.
        xvg_files : list of str
            The `xvg` files to be generated. They contain the tabulated LJ-9-6 potential and force.
        '''
        if itp_files is None:
            itp_files = list(filter(lambda x: x.endswith('.itp'), os.listdir('.')))
        for itp in itp_files:
            GMX._modify_lj96_itp(itp)

        if top_files is None:
            top_files = list(filter(lambda x: x.endswith('.top'), os.listdir('.')))
        for top in top_files:
            GMX._modify_lj96_top(top)

        if mdp_files is None:
            mdp_files = list(filter(lambda x: x.endswith('.mdp'), os.listdir('.')))
        for mdp in mdp_files:
            GMX._modify_lj96_mdp(mdp)

        for xvg in xvg_files:
            shutil.copy(os.path.join(GMX._TEMPLATE_DIR, 'table6-9.xvg'), xvg)

    @staticmethod
    def _modify_lj96_itp(itp):
        with open(itp) as f:
            lines = f.read().splitlines()
        new_lines = []
        for line in lines:
            if line.strip().split()[:3] == ['1', '2', 'yes']:
                new_lines.append(line + ' 9')
            else:
                new_lines.append(line)
        with open(itp, 'w') as f:
            f.write('\n'.join(new_lines))

    @staticmethod
    def _modify_lj96_top(top):
        with open(top) as f:
            lines = f.read().splitlines()

        new_lines = []
        ATOMLINE = False
        for line in lines:
            if line.startswith(';') or line.strip() == '':
                new_lines.append(line)
                continue

            if line.find('[') != -1 and line.find('atoms') != -1:
                ATOMLINE = True
                new_lines.append(line)
                continue

            if ATOMLINE and line.find('[') != -1:
                ATOMLINE = False

            if ATOMLINE:
                words = line.strip().split()
                words[5] = words[0]
                tmp = ''
                for word in words:
                    tmp += ' %10s' % word
                new_lines.append(tmp)
            else:
                new_lines.append(line)
                continue

        with open(top, 'w') as f:
            f.write('\n'.join(new_lines))

    @staticmethod
    def _modify_lj96_mdp(mdp):
        with open(mdp) as f:
            lines = f.read().splitlines()
        new_lines = []
        for line in lines:
            if line.strip().startswith('cutoff-scheme') or \
                    line.strip().startswith('cutoff_scheme') or \
                    line.strip().startswith('cutoffscheme') or \
                    line.strip().startswith('rlist'):
                pass
            else:
                new_lines.append(line)

            if line.strip().startswith('rvdw'):
                rvdw = line.split('=')[-1].strip()

        new_lines.append('\n; vdw table')
        new_lines.append('cutoff-scheme = group')
        new_lines.append('rlist = ' + rvdw)
        new_lines.append('vdwtype = user')
        with open(mdp, 'w') as f:
            f.write('\n'.join(new_lines))



from mstk.forcefield.forcefield import ForceField
from mstk.forcefield.ffterm import *
from mstk.chem.element import Element
from mstk.forcefield.errors import *
from mstk import logger


class PpfLine:
    def __init__(self, term, key, value, comment):
        self.term = term.strip()
        self.key = key.strip()
        self.value = value.strip()
        self.comment = comment.strip()
        self.name = self.term + '-' + self.key

        version = self.comment.split(',')[0].strip()
        if not version.startswith('V'):
            version = None
        else:
            version = version.lstrip('V').strip()
            try:
                float(version)
            except:
                raise Exception('Invalid version information: %s' % comment)
        self.version = version

    def get_names(self):
        names = [x.strip() for x in self.key.split(',')]
        for i in range(len(names)):
            if names[i] == 'X':
                names[i] = '*'
        return names

    def get_float_values(self):
        return [float(x.strip().strip('*')) for x in self.value.split(',')]


class Ppf:
    '''
    Parse ForceField from PPF file.

    PPF is the native format of `Direct Force Field`.
    Currently, only the DFF-AMBER functional form of PPF is supported,
    which contains following terms:

    * :class:`~mstk.forcefield.AtomType` with equivalent tables, masses and partial charges
    * :class:`~mstk.forcefield.ChargeIncrementTerm`
    * :class:`~mstk.forcefield.HarmonicBondTerm`
    * :class:`~mstk.forcefield.HarmonicAngleTerm`
    * :class:`~mstk.forcefield.PeriodicDihedralTerm`
    * :class:`~mstk.forcefield.PeriodicImproperTerm`

    In PPF file, there is no leading 1/2 for harmonic energy terms.
    The length is in unit of Angstrom, and angle is in unit of degree.
    The energies for bond, angle and vdW are in unit of kcal/mol/A^2, kcal/mol/rad^2 and kcal/mol.
    The two values for LJ interaction are r_min and epsilon.

    PPF format can optionally store the version of parameters.
    If a force field term describing one topology element appear several times,
    only the one with the latest version will be loaded.
    If there are duplicated terms without version record, an Exception will be raised.

    Parameters
    ----------
    file : str

    Attributes
    ----------
    forcefield : ForceField
    '''

    def __init__(self, file):
        ff = ForceField()
        ff.lj_mixing_rule = ForceField.LJ_MIXING_LB
        ff.vdw_long_range = ForceField.VDW_LONGRANGE_CORRECT
        ff.vdw_cutoff = 1.2  # nm
        ff.scale_14_vdw = 0.5
        ff.scale_14_coulomb = 1.0 / 1.2

        # save all lines to this so that we can keep only the latest version
        self._ppf_lines = {}
        self._parse(file)
        self._setup(ff)

        self.forcefield = ff

    def _parse(self, file):
        with open(file) as f:
            lines = f.read().splitlines()

        _section = ''
        for line in lines:
            line = line.strip()
            if line.startswith('#DFF:EQT'):
                _section = 'EQT'
                continue
            elif line.startswith('#PROTOCOL'):
                protocol = line.split('=')[-1].strip()
                if protocol != 'AMBER':
                    raise NotImplementedError('Current only AMBER protocol is implemented for PPF')
                continue
            elif line.startswith('#DFF:PPF'):
                _section = 'PPF'
                continue
            elif line.startswith('#') or line == '':
                continue
            if _section == '':
                raise Exception('Invalid PPF file')

            self._parse_line(line, _section)

    def _parse_line(self, line, section):
        words = [x.strip() for x in line.strip().split(':')]
        if section == 'EQT':
            term = 'EQT'
            key, value = words[:2]
            comment = words[2] if len(words) > 2 else ''
        elif section == 'PPF':
            term, key, value = words[:3]
            comment = words[3] if len(words) > 3 else ''
        else:
            raise Exception('Invalid section: %s' % section)

        ppf_line = PpfLine(term, key, value, comment)
        # only keep the line of the latest version
        old_line = self._ppf_lines.get(ppf_line.name)
        if old_line is None:
            self._ppf_lines[ppf_line.name] = ppf_line
        else:
            if ppf_line.version is None or old_line.version is None:
                raise Exception('Duplicated line without version information: %s' % ppf_line.name)
            if float(ppf_line.version) > float(old_line.version):
                self._ppf_lines[ppf_line.name] = ppf_line

    def _setup(self, ff):
        for line in self._ppf_lines.values():
            if line.term != 'EQT':
                continue
            atype = AtomType(line.key)
            # _eqt_charge_ is a temporary variable for matching charge
            (atype.eqt_vdw, atype._eqt_charge_, atype.eqt_q_inc, atype.eqt_bond,
             atype.eqt_ang_c, atype.eqt_ang_s, atype.eqt_dih_c, atype.eqt_dih_s,
             atype.eqt_imp_c, atype.eqt_imp_s) = line.value.split()
            atype.version = line.version
            ff.atom_types[atype.name] = atype

        for line in self._ppf_lines.values():
            if line.term == 'ATYPE':
                atype = ff.atom_types.get(line.key)
                if atype is None:
                    continue
                number, mass = line.get_float_values()
                atype.mass = mass
            elif line.term == 'ATC':
                for atype in ff.atom_types.values():
                    if atype._eqt_charge_ == line.key:
                        atype.charge = line.get_float_values()[0]
            elif line.term == 'BINC':
                try:
                    term = ChargeIncrementTerm(*line.get_names(), *line.get_float_values())
                except ChargeIncrementNonZeroError:
                    logger.warning('Invalid BINC line ignored: %s' % line.key)
                else:
                    term.version = line.version
                    ff.qinc_terms[term.name] = term
            elif line.term == 'N12_6':
                at = line.get_names()[0]
                r_min, epsilon = line.get_float_values()
                term = LJ126Term(at, at, epsilon * 4.184, r_min / 2 ** (1 / 6) / 10)
                term.version = line.version
                ff.vdw_terms[term.name] = term
            elif line.term == 'P12_6':
                r_min, epsilon = line.get_float_values()
                term = LJ126Term(*line.get_names(), epsilon * 4.184, r_min / 2 ** (1 / 6) / 10)
                term.version = line.version
                ff.pairwise_vdw_terms[term.name] = term
            elif line.term == 'BHARM':
                length, k = line.get_float_values()
                at1, at2 = line.get_names()
                # constrain the bonds involving hydrogen
                fixed = at1.startswith('h_1') or at2.startswith('h_1')
                term = HarmonicBondTerm(at1, at2, length / 10, k * 4.184 * 100, fixed=fixed)
                term.version = line.version
                ff.bond_terms[term.name] = term
            elif line.term == 'AHARM':
                theta, k = line.get_float_values()
                at1, at2, at3 = line.get_names()
                # constrain the angles in water
                fixed = at1.startswith('h_1') and at2.startswith('o_2') and at3.startswith('h_1')
                term = HarmonicAngleTerm(at1, at2, at3, theta * DEG2RAD, k * 4.184, fixed=fixed)
                term.version = line.version
                ff.angle_terms[term.name] = term
            elif line.term == 'TCOSP':
                values = line.get_float_values()
                count = len(values) // 3
                term = PeriodicDihedralTerm(*line.get_names())
                for i in range(count):
                    phi, k, n = values[i * 3], values[i * 3 + 1] * 4.184, int(values[i * 3 + 2])
                    # ignore nonsense values
                    if n == 0 and k == 0:
                        continue
                    term.add_phase(phi * DEG2RAD, k, n)
                term.version = line.version
                ff.dihedral_terms[term.name] = term
            elif line.term == 'IBCOS':
                # the third atom is the center atom for IBCOS improper term
                a1, a2, a3, a4 = line.get_names()
                try:
                    phi, k, n = line.get_float_values()
                except:
                    raise Exception('IBCOS should contain three parameters')
                if phi != 180 or n != 2:
                    raise Exception('IBCOS should have phi=180 and n=2: %s' % line.key)
                term = OplsImproperTerm(a3, a1, a2, a4, k * 4.184)
                term.version = line.version
                ff.improper_terms[term.name] = term

        # update the mass based on eqt_vdw
        for atype in ff.atom_types.values():
            if atype.eqt_vdw != atype.name:
                atype_eq = ff.atom_types.get(atype.eqt_vdw)
                if atype_eq is None:
                    raise Exception('vdW equivalent type for %s not found in PPF' % (str(atype)))
                atype.mass = atype_eq.mass


ForceField.registor_format('.ppf', Ppf)



import xml.etree.ElementTree as ET
from xml.dom import minidom
from mstk.forcefield.forcefield import ForceField
from mstk.forcefield.ffterm import *
from mstk import logger


class Zfp:
    '''
    Load ForceField from ZFP file.

    ZFP is the default format for storing ForceField in mstk.
    XML language is used by ZFP file to serialize the setting and all the FFTerms in a ForceField object.

    In ZFP file, there is no leading 1/2 for harmonic energy terms.
    The length is in unit of nm, and angle is in unit of degree.
    The energies for harmonic bond, harmonic angle and vdW are in unit of kJ/mol/nm^2, kJ/mol/rad^2 and kJ/mol.

    A force field term describing one topology element should only appear once.
    If there are duplicated terms, an Exception will be raised.
    e.g. HarmonicAngleTerm('c_4', 'c_4', 'h_1') and SDKAngleTerm('c_4', 'c_4', 'h_1') are considered as duplicated,
    because both of them are describing the angle type ('c_4', 'c_4', 'h_1').
    LJ126Term('c_4', 'h_1') and MieTerm('c_4', 'h_1') are also duplicated,
    because both of them are describing the vdW interactions between atom types 'c_4' and 'h_1'.

    TODO Because ZFP is designed mainly for FF parameters exchange, several features are not currently supported by ZFP format
    1. Comments for force field and ffterm
    2. Adjustable parameters

    Parameters
    ----------
    file : str

    Attributes
    ----------
    forcefield : ForceField
    '''

    def __init__(self, file):
        self.forcefield = ForceField()
        self._parse(file)

    def _parse(self, file):
        try:
            tree = ET.ElementTree(file=file)
        except:
            raise Exception('Invalid ZFP file')

        root = tree.getroot()
        if root is None:
            raise Exception('Empty ZFP file')

        ff = self.forcefield

        node = root.find('Setting')
        ff.vdw_cutoff = float(node.attrib['vdw_cutoff'])
        ff.vdw_long_range = node.attrib['vdw_long_range']
        ff.lj_mixing_rule = node.attrib['lj_mixing_rule']
        ff.scale_14_vdw = float(node.attrib['scale_14_vdw'])
        ff.scale_14_coulomb = float(node.attrib['scale_14_coulomb'])

        tags = {
            'AtomTypes'           : ff.atom_types,
            'VirtualSiteTerms'    : ff.virtual_site_terms,
            'ChargeIncrementTerms': ff.qinc_terms,
            'VdwTerms'            : ff.vdw_terms,
            'PairwiseVdwTerms'    : ff.pairwise_vdw_terms,
            'BondTerms'           : ff.bond_terms,
            'AngleTerms'          : ff.angle_terms,
            'DihedralTerms'       : ff.dihedral_terms,
            'ImproperTerms'       : ff.improper_terms,
            'PolarizableTerms'    : ff.polarizable_terms,
        }
        _duplicated = []
        for tag, d in tags.items():
            node = root.find(tag)
            if node is None:
                continue
            for element in node:
                try:
                    term = FFTermFactory.create_from_zfp(element.tag, element.attrib)
                except:
                    raise Exception('Invalid tag or attributes: %s, %s' %
                                    (element.tag, str(element.attrib)))
                if term.name in d.keys():
                    _duplicated.append(term)
                else:
                    d[term.name] = term

        if _duplicated:
            msg = f'{len(_duplicated)} duplicated terms from {file}'
            for term in _duplicated:
                msg += f'\n        {term}'
            logger.error(msg)
            raise Exception('Duplicated terms in ZFP file')

    @staticmethod
    def save_to(ff, file):
        '''
        Save ForceField to a ZFP file

        Parameters
        ----------
        ff : ForceField
        file : str
        '''
        root = ET.Element('ForceFieldTerms')

        attrib = {
            'vdw_cutoff'      : str(ff.vdw_cutoff),
            'vdw_long_range'  : ff.vdw_long_range,
            'lj_mixing_rule'  : ff.lj_mixing_rule,
            'scale_14_vdw'    : str(ff.scale_14_vdw),
            'scale_14_coulomb': str(ff.scale_14_coulomb),
        }
        node = ET.SubElement(root, 'Setting', attrib=attrib)

        tags = {
            'AtomTypes'           : ff.atom_types,
            'VirtualSiteTerms'    : ff.virtual_site_terms,
            'ChargeIncrementTerms': ff.qinc_terms,
            'VdwTerms'            : ff.vdw_terms,
            'PairwiseVdwTerms'    : ff.pairwise_vdw_terms,
            'BondTerms'           : ff.bond_terms,
            'AngleTerms'          : ff.angle_terms,
            'DihedralTerms'       : ff.dihedral_terms,
            'ImproperTerms'       : ff.improper_terms,
            'PolarizableTerms'    : ff.polarizable_terms,
        }
        for tag, d in tags.items():
            node = ET.SubElement(root, tag)
            for term in d.values():
                ET.SubElement(node, term.__class__.__name__, attrib=term.to_zfp())

        str_xml = minidom.parseString(ET.tostring(root)).toprettyxml(indent='  ')
        with open(file, 'wb') as f:
            f.write(str_xml.encode())


ForceField.registor_format('.zfp', Zfp)



#!/usr/bin/env python3

import os
import argparse
import numpy as np


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-b', '--begin', type=float, default=0)
    parser.add_argument('-p', '--plot', action='store_true', help='plot the histogram')
    parser.add_argument('--skip', type=int, default=1)
    parser.add_argument('--nbin', type=int, help='number of bins for histogram')
    parser.add_argument('--type', type=str, default='openmm', choices=['openmm', 'lammps', 'plumed'])

    return parser.parse_args()


if __name__ == '__main__':
    args = parse_args()

    BEGIN = args.begin
    SKIP = args.skip

    r0_k_list = []
    files = filter(os.path.isdir, os.listdir(os.getcwd()))
    for i in files:
        if os.path.isdir(i) and '-' in i and \
                (i[0].isdigit() or i[0].startswith('_')) and i[-1].isdigit():
            r0 = i.split('-')[0]
            k = i.split('-')[1]
            r0_k_list.append((r0, k))
    r0_k_list.sort(key=lambda x: float(x[0].replace('_', '-')))

    outf_meta = open('wham-meta', 'w')
    for r0, k in r0_k_list:
        outfile = 'wham-' + r0 + '-' + k + '.txt'
        outf = open(outfile, 'w')
        if args.type == 'openmm':
            _START = False
            with open(r0 + '-' + k + '/_job.out') as f:
                lines = f.read().splitlines()
            for line in lines:
                if line.startswith('#"'):
                    _START = True
                    continue
                if not _START:
                    continue
                try:
                    words = line.strip().split()
                    step = int(words[0])
                except:
                    continue
                if step >= BEGIN and step % SKIP == 0:
                    str = line.strip().split()
                    time = float(words[0])
                    r = float(words[-1])
                    outf.write('%.1f %.6f\n' % (time, r))
        elif args.type == 'lammps':
            _CYCLE = 0
            _START = False
            countline = 0
            with open(r0 + '-' + k + '/log') as f:
                lines = f.read().splitlines()
            for line in lines:
                if line.startswith('Loop') and _CYCLE == 2:
                    break
                if line.startswith('Step'):
                    _CYCLE += 1
                    if _CYCLE == 2:
                        _START = True
                if _START:
                    countline += 1
                    if countline > BEGIN + 1 and countline % SKIP == 0:
                        str = line.strip().split()
                        time = float(str[0])
                        r = float(str[6])
                        outf.write('%.1f %.6f\n' % (time, r))
        elif args.type == 'plumed':
            countline = 0
            with open(r0 + '-' + k + '/DIS.txt') as f:
                lines = f.read().splitlines()
            for line in lines:
                countline += 1
                if countline > BEGIN + 1 and countline % SKIP == 0:
                    str = line.strip().split()
                    time = float(str[0])
                    r = float(str[1])
                    outf.write('%.1f %.6f\n' % (time, r))
        outf.close()
        outf_meta.write('%s %s %s\n' % (outfile, r0.replace('_', '-'), k))  # replace the leading _ with -
    outf_meta.close()

    if args.plot:
        import matplotlib

        matplotlib.use('Agg')
        import matplotlib.pyplot as plt

        fig = plt.gcf()
        fig.set_size_inches(16, 8)
        plt.grid(True)
        files = filter(lambda x: x.startswith('wham-') and x.endswith('.txt'), os.listdir(os.getcwd()))
        for r0, k in r0_k_list:
            array = np.genfromtxt('wham-%s-%s.txt' % (r0, k))
            data = array[:, 1]
            if args.nbin:
                hist, bins = np.histogram(data, bins=args.nbin)
            else:
                hist, bins = np.histogram(data)
            center = (bins[:-1] + bins[1:]) / 2.
            plt.plot(center, hist, linewidth=2)
        plt.savefig('wham-hist.png')

    # os.system('wham %f %f %s %s 298.15 0 wham-meta wham-pmf' %(hist_min, hist_max, bins, tol))



import itertools
from .system import System
from ..forcefield import *
from ..topology import *
from ..trajectory import *
from .. import logger


class GromacsExporter():
    '''
    GromacsExporter export a :class:`System` to input files for Gromacs.

    The following potential functions are currently supported:

    * :class:`~mstk.forcefield.LJ126Term`
    * :class:`~mstk.forcefield.HarmonicBondTerm`
    * :class:`~mstk.forcefield.HarmonicAngleTerm`
    * :class:`~mstk.forcefield.LinearAngleTerm`
    * :class:`~mstk.forcefield.OplsDihedralTerm`
    * :class:`~mstk.forcefield.PeriodicDihedralTerm`
    * :class:`~mstk.forcefield.OplsImproperTerm`
    * :class:`~mstk.forcefield.HarmonicImproperTerm`
    * :class:`~mstk.forcefield.DrudeTerm`

    The :class:`~mstk.forcefield.MieTerm` can be exported, but it will be in the LJ-12-6 form.
    If the dispersion of `MieTerm` is in 6-th power, then MieTerm can be correctly handled
    by small modifications to topol and mdp files after exported.
    Refer to the GROMACS documentation for details.

    The :class:`~mstk.forcefield.SDKAngleTerm` can be exported, but it will be in the harmonic form.
    Usually it is acceptable if the molecule is not going to form a coil structure.
    Refer to the original SDK paper for details.
    '''

    def __init__(self):
        pass

    @staticmethod
    def export(system, gro_out, top_out, mdp_out, **kwargs):
        '''
        Generate input files for Gromacs from a system

        # TODO Export virtual sites other than TIP4P

        Parameters
        ----------
        system : System
        gro_out : str or None
        top_out : str or None
        mdp_out : str or None
        '''
        if not system.use_pbc:
            raise Exception('PBC required for exporting GROMACS')

        supported_terms = {LJ126Term, MieTerm,
                           HarmonicBondTerm, MorseBondTerm,
                           HarmonicAngleTerm, SDKAngleTerm, LinearAngleTerm,
                           OplsDihedralTerm, PeriodicDihedralTerm,
                           OplsImproperTerm, HarmonicImproperTerm,
                           DrudeTerm}
        unsupported = system.ff_classes - supported_terms
        if unsupported != set():
            raise Exception('Unsupported FF terms: %s' % (', '.join(map(lambda x: x.__name__, unsupported))))

        if MieTerm in system.ff_classes:
            logger.warning('MieTerm not supported by GROMACS. Exported in LJ-12-6 form')
        if SDKAngleTerm in system.ff_classes:
            logger.warning('SDKAngleTerm not supported by GROMACS. Exported in harmonic form')
        if DrudeTerm in system.ff_classes:
            logger.warning('DrudeTerm not well tested in GROMACS. Use it with caution')

        if system.vsite_types - {TIP4PSite} != set():
            raise Exception('Virtual sites other than TIP4PSite haven\'t been implemented')

        if gro_out is not None:
            GromacsExporter._export_gro(system, gro_out)
        if top_out is not None:
            GromacsExporter._export_top(system, top_out)
        if mdp_out is not None:
            GromacsExporter._export_mdp(system, mdp_out)

    @staticmethod
    def _export_gro(system: System, gro_out='conf.gro'):
        frame = Frame(system._topology.n_atom)
        frame.cell = system._topology.cell
        frame.positions = system._topology.positions
        gro = Trajectory(gro_out, 'w')
        gro.write_frame(frame, system._topology)
        gro.close()

    @staticmethod
    def _export_top(system: System, top_out='topol.top'):
        mols_unique = system._topology.get_unique_molecules(deepcopy=False)

        string = '; GROMACS topol file created by mstk\n'

        string += '\n[ defaults ]\n'
        string += '; nbfunc  comb-rule  gen-pairs  fudgeLJ  fudgeQQ\n'
        _comb = 2 if system.ff.lj_mixing_rule == ForceField.LJ_MIXING_LB else 3
        string += '%6i %6i %12s %10.4f %10.4f\n' % (
            1, _comb, 'yes', system.ff.scale_14_vdw, system.ff.scale_14_coulomb)

        string += '\n[ atomtypes ]\n'
        string += ';     name       mass     charge      ptype      sigma      epsilon\n'
        drude_types = set()
        if DrudeTerm in system.ff_classes:
            for atom in system.topology.atoms:
                if atom.is_drude:
                    drude_types.add(atom.type)
        for atype in system.ff.atom_types.values():
            ptype = 'S' if atype.name in drude_types else 'A'
            vdw = system.ff.get_vdw_term(atype, atype)
            if vdw.__class__ in (LJ126Term, MieTerm):
                string += '%10s %10.4f %12.6f %6s %12.6f %12.6f  ; %s\n' % (
                    atype.name, 0.0, 0.0, ptype, vdw.sigma, vdw.epsilon, ' '.join(vdw.comments))
            else:
                raise Exception('Unsupported vdW term')

        string += '\n[ nonbond_params ]\n'
        string += ';       i        j        func      sigma      epsilon\n'
        for at1, at2 in itertools.combinations(system.ff.atom_types.values(), 2):
            try:
                vdw = system.ff.get_vdw_term(at1, at2, mixing=False)
            except:
                continue
            if vdw.__class__ in (LJ126Term, MieTerm):
                string += '%10s %10s %6i %12.6f %12.6f  ; %s\n' % (
                    at1.name, at2.name, 1, vdw.sigma, vdw.epsilon, ' '.join(vdw.comments))
            else:
                raise Exception('Unsupported vdW term')

        for i, mol in enumerate(mols_unique.keys()):
            mol: Molecule
            string += '\n[ moleculetype ]\n'
            string += '; Name            nrexcl\n'
            string += f'{mol.name}_{i + 1}     3\n'

            string += '\n[ atoms ]\n'
            string += ';%5s %10s %6s %10s %6s %6s %12s %10s\n' % (
                'nr', 'type', 'resnr', 'residue', 'atom', 'cgnr', 'charge', 'mass')
            for j, atom in enumerate(mol.atoms):
                # GROMACS use SCF approach for Drude simulation. Set mass of Drude to zero
                mass = 0 if atom.is_drude else atom.mass
                if atom in system.drude_pairs:
                    mass += system.drude_pairs[atom].mass
                string += '%6i %10s %6i %10s %6s %6i %12.6f %10.4f\n' % (
                    j + 1, atom.type, atom.residue.id_in_mol + 1, atom.residue.name,
                    atom.symbol, j + 1, atom.charge, mass)

            if TIP4PSite in system.vsite_types:
                string += '\n[ virtual_sites3 ]\n'
                string += ';%5s %6s %6s %6s %6s %12s %12s\n' % ('Vsite', 'from1', 'from2', 'from3', 'funct', 'a', 'b')
                for atom in mol.atoms:
                    vsite = atom.virtual_site
                    if type(vsite) is TIP4PSite:
                        O, H1, H2 = vsite.parents
                        w_O, w_H1, w_H2 = system.get_TIP4P_linear_coeffs(atom)
                        string += '%6i %6i %6i %6i %6i %12.6f %12.6f\n' % (
                            atom.id_in_mol + 1, O.id_in_mol + 1, H1.id_in_mol + 1, H2.id_in_mol + 1, 1, w_H1, w_H2)
                    elif vsite is not None:
                        raise Exception('Virtual sites other than TIP4PSite haven\'t been implemented')

            string += '\n[ pairs ]\n'
            pairs12, pairs13, pairs14 = mol.get_12_13_14_pairs()
            for a1, a2 in pairs14:
                string += '%6i %6i %6i\n' % (a1.id_in_mol + 1, a2.id_in_mol + 1, 1)

            # 1-4 pairs concerning Drude particles
            drude_pairs = dict(mol.get_drude_pairs())
            _pairs14_drude = []
            for a1, a2 in pairs14:
                if a1 in drude_pairs:
                    _pairs14_drude.append((drude_pairs[a1], a2))
                if a2 in drude_pairs:
                    _pairs14_drude.append((a1, drude_pairs[a2]))
                if a1 in drude_pairs and a2 in drude_pairs:
                    _pairs14_drude.append((drude_pairs[a1], drude_pairs[a2]))

            # 1-4 pairs concerning virtual sites
            vsite_pairs = dict(mol.get_virtual_site_pairs())
            _pairs14_vsite = []
            for atom1, atom2 in pairs14:
                vsite1 = vsite_pairs.get(atom1)
                vsite2 = vsite_pairs.get(atom2)
                drude1 = drude_pairs.get(atom1)
                drude2 = drude_pairs.get(atom2)
                if vsite1 is not None:
                    _pairs14_vsite.append((vsite1, atom2))
                    if drude2 is not None:
                        _pairs14_vsite.append((vsite1, drude2))
                if vsite2 is not None:
                    _pairs14_vsite.append((vsite2, atom1))
                    if drude1 is not None:
                        _pairs14_vsite.append((vsite2, drude1))
                if None not in [vsite1, vsite2]:
                    _pairs14_vsite.append((vsite1, vsite2))

            for a1, a2 in _pairs14_drude + _pairs14_vsite:
                string += '%6i %6i %6i\n' % (a1.id_in_mol + 1, a2.id_in_mol + 1, 1)

            ### Drude polarization #############################################
            if DrudeTerm in system.ff_classes:
                string += '\n[ polarization ]\n'
                string += ';  ai    aj   func    alpha     delta     khyp\n'
                for parent, drude in drude_pairs.items():
                    string += '%5d %5d   %4d  %9.6f  %7.4f  %9.4e\n' % (
                        parent.id_in_mol + 1, drude.id_in_mol + 1,
                        2, parent.alpha, 0.02, 16.768e8)

                string += '\n[ thole_polarization ]\n'
                string += ';  ai    aj    ak    al   func   thole   alpha1   alpha2\n'
                for a1, a2 in pairs12 + pairs13:
                    if a1 in drude_pairs and a2 in drude_pairs:
                        string += '%5i %5i %5i %5i   %4i   %9.5f  %9.6f  %9.6f\n' % (
                            a1.id_in_mol + 1, drude_pairs[a1].id_in_mol + 1,
                            a2.id_in_mol + 1, drude_pairs[a2].id_in_mol + 1,
                            1, (a1.thole + a2.thole) / 2, a1.alpha, a2.alpha)

            string += '\n[ constraints ]\n'
            for bond in mol.bonds:
                distance = system.constrain_bonds.get(bond)
                if distance is not None:
                    a1, a2 = bond.atom1, bond.atom2
                    string += '%6i %6i %6i %12.6f\n' % (
                        a1.id_in_mol + 1, a2.id_in_mol + 1, 1, distance)
            for angle in mol.angles:
                distance = system.constrain_angles.get(angle)
                if distance is not None:
                    a1, a3 = angle.atom1, angle.atom3
                    string += '%6i %6i %6i %12.6f\n' % (
                        a1.id_in_mol + 1, a3.id_in_mol + 1, 2, distance)

            string += '\n[ bonds ]\n'
            for bond in mol.bonds:
                if bond.is_drude:
                    continue
                if bond in system.constrain_bonds:
                    continue
                bterm = system.bond_terms[bond]
                if bterm.__class__ == HarmonicBondTerm:
                    a1, a2 = bond.atom1, bond.atom2
                    string += '%6i %6i %6i %12.6f %12.4f\n' % (
                        a1.id_in_mol + 1, a2.id_in_mol + 1,
                        1, bterm.length, bterm.k * 2)
                elif bterm.__class__ == MorseBondTerm:
                    a1, a2 = bond.atom1, bond.atom2
                    string += '%6i %6i %6i %12.6f %12.4f %12.4f\n' % (
                        a1.id_in_mol + 1, a2.id_in_mol + 1,
                        3, bterm.length, bterm.depth, (bterm.k / bterm.depth) ** 0.5)
                else:
                    raise Exception('Unsupported bond term')

            # Exclusions concerning Drude particles
            _exclusions_drude = {}  # {drude1: [atom2, drude2, atom3, drude3, ...]}
            for a1, a2 in pairs12 + pairs13 + pairs14:
                if a1 in drude_pairs:
                    d1 = drude_pairs[a1]
                    if d1 not in _exclusions_drude:
                        _exclusions_drude[d1] = []
                    _exclusions_drude[d1].append(a2)
                if a2 in drude_pairs:
                    d2 = drude_pairs[a2]
                    if d2 not in _exclusions_drude:
                        _exclusions_drude[d2] = []
                    _exclusions_drude[d2].append(a1)
                if a1 in drude_pairs and a2 in drude_pairs:
                    _exclusions_drude[d1].append(d2)

            # Exclusions concerning virtual sites
            _exclusions_vsite = {}  # {avsite1: [atom1, drude1, atom2, drude2, avsite2, ...]}
            for atom, vsite in vsite_pairs.items():
                _exclusions_vsite[vsite] = [atom]
                drude = drude_pairs.get(atom)
                if drude is not None:
                    _exclusions_vsite[vsite].append(drude)
            for atom1, atom2 in pairs12 + pairs13:
                vsite1 = vsite_pairs.get(atom1)
                vsite2 = vsite_pairs.get(atom2)
                drude1 = drude_pairs.get(atom1)
                drude2 = drude_pairs.get(atom2)
                if vsite1 is not None:
                    _exclusions_vsite[vsite1].append(atom2)
                    if drude2 is not None:
                        _exclusions_vsite[vsite1].append(drude2)
                if vsite2 is not None:
                    _exclusions_vsite[vsite2].append(atom1)
                    if drude1 is not None:
                        _exclusions_vsite[vsite2].append(drude1)
                if None not in [vsite1, vsite2]:
                    _exclusions_vsite[vsite1].append(vsite2)

            _exclusions = {**_exclusions_drude, **_exclusions_vsite}
            if len(_exclusions) > 0:
                string += '\n[ exclusions ]\n'
                string += ';  ai    aj    ...\n'
                for a1, a2_list in _exclusions.items():
                    string += '%5d' % (a1.id_in_mol + 1)
                    for a2 in a2_list:
                        string += ' %5d' % (a2.id_in_mol + 1)
                    string += '\n'

            string += '\n[ angles ]\n'
            for angle in mol.angles:
                if angle in system.constrain_angles:
                    continue
                aterm = system.angle_terms[angle]
                a1, a2, a3 = angle.atom1, angle.atom2, angle.atom3
                if aterm.__class__ in (HarmonicAngleTerm, SDKAngleTerm):
                    string += '%6i %6i %6i %6i %12.6f %12.4f\n' % (
                        a1.id_in_mol + 1, a2.id_in_mol + 1, a3.id_in_mol + 1,
                        1, aterm.theta * RAD2DEG, aterm.k * 2)
                elif aterm.__class__ is LinearAngleTerm:
                    bond12, bond23 = angle.bonds
                    bterm12 = system.bond_terms[bond12]
                    bterm23 = system.bond_terms[bond23]
                    a, k_linear = aterm.calc_a_k_linear(bterm12.length, bterm23.length)
                    string += '%6i %6i %6i %6i %12.6f %12.4f\n' % (
                        a1.id_in_mol + 1, a2.id_in_mol + 1, a3.id_in_mol + 1,
                        9, a, k_linear * 2)
                else:
                    raise Exception('Unsupported angle term')

            string += '\n[ dihedrals ]\n'
            for dihedral in mol.dihedrals:
                dterm = system.dihedral_terms[dihedral]
                if dterm.__class__ == OplsDihedralTerm:
                    dterm = dterm.to_periodic_term()
                a1, a2, a3, a4 = dihedral.atom1, dihedral.atom2, dihedral.atom3, dihedral.atom4
                if dterm.__class__ == PeriodicDihedralTerm:
                    for phase in dterm.phases:
                        string += '%6i %6i %6i %6i %6i %8.2f %12.4f %4i\n' % (
                            a1.id_in_mol + 1, a2.id_in_mol + 1,
                            a3.id_in_mol + 1, a4.id_in_mol + 1,
                            9, phase.phi * RAD2DEG, phase.k, phase.n)
                    if len(dterm.phases) == 0:
                        string += '%6i %6i %6i %6i %6i %8.2f %12.4f %4i  ; no dihedral parameters\n' % (
                            a1.id_in_mol + 1, a2.id_in_mol + 1,
                            a3.id_in_mol + 1, a4.id_in_mol + 1,
                            9, 0.0, 0.0, 1)
                else:
                    raise Exception('Unsupported dihedral term')

            string += '\n[ dihedrals ]\n'
            for improper in mol.impropers:
                iterm = system.improper_terms[improper]
                a1, a2, a3, a4 = improper.atom1, improper.atom2, improper.atom3, improper.atom4
                if iterm.__class__ == OplsImproperTerm:
                    # be careful about the sequence of atoms in OPLS improper definition
                    string += '%6i %6i %6i %6i %6i %8.2f %12.4f %4i\n' % (
                        a2.id_in_mol + 1, a3.id_in_mol + 1,
                        a1.id_in_mol + 1, a4.id_in_mol + 1,
                        4, 180, iterm.k, 2)
                elif iterm.__class__ == HarmonicImproperTerm:
                    string += '%6i %6i %6i %6i %6i %8.2f %12.4f\n' % (
                        a1.id_in_mol + 1, a2.id_in_mol + 1,
                        a3.id_in_mol + 1, a4.id_in_mol + 1,
                        2, iterm.phi * RAD2DEG, iterm.k * 2)
                else:
                    raise Exception('Unsupported improper term')

        string += '\n[ system ]\n'

        string += '\n[ molecules ]\n'
        for i, (mol, count) in enumerate(mols_unique.items()):
            string += '%s_%i %6i\n' % (mol.name, i + 1, count)

        with open(top_out, 'wb') as f:
            f.write(string.encode())

    @staticmethod
    def _export_mdp(system: System, mdp_out='grompp.mdp'):
        r_cut = system.ff.vdw_cutoff
        tau_t = 0.2 if DrudeTerm in system.ff_classes else 1.0
        string = f'''; Created by mstk
integrator      = sd
dt              = 0.002 ; ps
nsteps          = 1000000

nstxout         = 0
nstvout         = 0
nstfout         = 0
nstxout-compressed = 1000
compressed-x-grps  = System

cutoff-scheme   = Verlet
pbc             = xyz
; rlist           = {r_cut}
coulombtype     = PME
rcoulomb        = {r_cut}
vdwtype         = Cut-off
rvdw            = {r_cut}
DispCorr        = EnerPres

tcoupl          = no; v-rescale
tc_grps         = System
tau_t           = {tau_t}
ref_t           = 300

pcoupl          = berendsen ; parrinello-rahman
pcoupltype      = isotropic
tau_p           = 0.5; 5
compressibility = 4.5e-5
ref_p           = 1.0

gen_vel         = yes
gen_temp        = 300

constraints     = h-bonds
constraint-algorithm = LINCS
'''
        with open(mdp_out, 'wb')  as f:
            f.write(string.encode())



from .frame import Frame
from .trajectory import Trajectory
from .handler import TrjHandler
from .io.gro import Gro
from .io.dcd import Dcd
from .io.lammps import LammpsTrj
from .io.xtc import Xtc
from .io.xyz import Xyz
from .io.combined_trj import CombinedTrj



import math
import numpy as np


class UnitCell:
    '''
    UnitCell represents the periodic boundary condition of simulation system.

    Both rectangular and triclinic unit cells are supported by this class.
    However, almost all other modules in `mstk` only support rectangular cell.

    Parameters
    ----------
    box : array_like, optional
        box can be array_like of shape (3,), (2,3) or (3,3).
        If its shape is (3,), then it will be treated as the lengths of a rectangular box.
        If its shape is (2,3), then it will be treated as the lengths and angles of a triclinic box.
        The angles are in unit of radian.
        If its shape is (3,3), then it will be treated as the box vectors.
        If not provided, the cell will be initialized with all elements of box vectors equal to zero.
    '''

    def __init__(self, box=None):
        self._vectors = None
        self._lengths = None
        self._angles = None
        self.set_box(box)

    def set_box(self, box):
        '''
        Set the unit cell with lengths or (lengths and angles) or full vectors.

        Parameters
        ----------
        box : array_like or None
            See the argument `box` of constructor for detailed explanations.
        '''
        if box is None:
            if self._vectors is None:
                self._vectors = np.zeros((3, 3), dtype=float)
            else:
                self._vectors.fill(0.0)
            if self._lengths is None:
                self._lengths = np.zeros(3, dtype=float)
            else:
                self._lengths.fill(0.0)
            if self._angles is None:
                self._angles = np.empty(3, dtype=float)
            self._angles.fill(math.pi / 2)
            return

        if not isinstance(box, (tuple, list, np.ndarray)):
            raise ValueError('Invalid argument')
        array = np.array(box, dtype=float)
        if array.shape == (3,):
            self._vectors = np.array([[array[0], 0, 0],
                                      [0, array[1], 0],
                                      [0, 0, array[2]]], dtype=float)
            self._lengths = array
            self._angles = np.array([math.pi / 2] * 3, dtype=float)
        elif array.shape == (3, 3):
            if array[0][1] != 0 or array[0][2] != 0 or array[1][2] != 0:
                raise ValueError('Invalid value for box vectors')
            self._reduce_box_vectors(array)
            self._vectors = array
            self._lengths = None  # calculate lengths and angles when required
            self._angles = None  # calculate lengths and angles when required
        elif array.shape == (2, 3):
            self._vectors = None  # calculate vectors when required
            self._lengths = array[0]
            self._angles = array[1]

        else:
            raise ValueError('Invalid argument')

    @property
    def vectors(self):
        '''
        The box vectors of the unit cell.

        The box vectors is a 3*3 matrix [[Vxx, Vxy, Vxz], [Vyx, Vyy, Vyz], [Vzx, Vzy, Vzz]].
        Several criterion will be satisfied:
        Vxy = Vxz = Vyz = 0; Vxx > 2*Vyx; Vxx > 2*Vzx; Vyy > 2*Vzy.

        Returns
        -------
        vectors : array_like
            The shape of box vectors is (3,3)
        '''
        if self._vectors is None:
            self._vectors = self.calc_vectors_from_lengths_angles(self._lengths, self._angles)
        return self._vectors

    @property
    def size(self):
        '''
        The diagonal values of the box vectors.

        Returns
        -------
        size : array_like
            The shape of the size vector is (3,)

        '''
        return np.array([self.vectors[0][0], self.vectors[1][1], self.vectors[2][2]])

    @property
    def volume(self):
        '''
        The volume of the unit cell.

        Returns
        -------
        volume : float
        '''
        return self.vectors[0][0] * self.vectors[1][1] * self.vectors[2][2]

    @property
    def lengths(self):
        '''
        The lengths of three edges of the unit cell.

        Returns
        -------
        lengths : array_like
            The shape of the length vector is (3,)

        '''
        if self._lengths is None:
            self._lengths, self._angles = self.calc_lengths_angles_from_vectors(self.vectors)
        return self._lengths

    @property
    def angles(self):
        '''
        The angles formed between the three edges of the unit cell, in unit of radian.

        Returns
        -------
        angles: array_like
            The shape of the angle vector is (3,)
        '''

        if self._angles is None:
            self._lengths, self._angles = self.calc_lengths_angles_from_vectors(self.vectors)
        return self._angles

    @property
    def is_rectangular(self):
        '''
        Whether or not this unit cell is rectangular.

        Returns
        -------
        is : bool
        '''
        return all(np.abs(self.angles - math.pi / 2) < 1e-4)

    @staticmethod
    def _reduce_box_vectors(vectors):
        '''
        Transform the orientations so that a0 >= 2*b0, a0 >= 2*c0, b1 >= 2*c1
        '''
        a, b, c = vectors
        if a[0] >= 2 * b[0] and a[0] >= 2 * c[0] and b[1] >= 2 * c[1]:
            return
        c -= b * round(c[1] / b[1])
        c -= a * round(c[0] / a[0])
        b -= a * round(b[0] / a[0])

    @staticmethod
    def calc_vectors_from_lengths_angles(lengths, angles):
        '''
        Convert box lengths and angles to periodic box vectors.

        Parameters
        ----------
        lengths : array_like
            Length vector in shape of (3,)
        angles : array_like
            Angle vector in shape of (3,) in unit o radian

        Returns
        -------
        vectors : array_like
            Box vectors in shape of (3,3)
        '''
        la, lb, lc = lengths
        alpha, beta, gamma = angles

        if all(angles - math.pi / 2 < 1e-4):
            return np.array([[la, 0, 0],
                             [0, lb, 0],
                             [0, 0, lc]], dtype=float)

        a = np.array([la, 0, 0], dtype=float)
        b = np.array([lb * math.cos(gamma), lb * math.sin(gamma), 0], dtype=float)
        cx = lc * math.cos(beta)
        cy = lc * (math.cos(alpha) - math.cos(beta) * math.cos(gamma)) / math.sin(gamma)
        cz = math.sqrt(lc * lc - cx * cx - cy * cy)
        c = np.array([cx, cy, cz], dtype=float)

        # if any element is very close to 0, set it to exactly 0
        for i in range(3):
            if abs(a[i]) < 1e-4:
                a[i] = 0.0
            if abs(b[i]) < 1e-4:
                b[i] = 0.0
            if abs(c[i]) < 1e-4:
                c[i] = 0.0

        # make sure ax > 2*bx, ax > 2*cz, by > 2*bz
        c = c - b * round(c[1] / b[1])
        c = c - a * round(c[0] / a[0])
        b = b - a * round(b[0] / a[0])

        return np.array([a, b, c])

    @staticmethod
    def calc_lengths_angles_from_vectors(vectors):
        '''
        Convert periodic box vectors to lengths and angles.

        Parameters
        ----------
        vectors : array_like
            Box vectors in shape of (3,3)

        Returns
        -------
        lengths : array_like
            Length vector in shape of (3,)
        angles : array_like
            Angle vector in shape of (3,) in unit of radian
        '''
        a, b, c = vectors
        la = np.sqrt(np.dot(a, a))
        lb = np.sqrt(np.dot(b, b))
        lc = np.sqrt(np.dot(c, c))

        if b[1] == 0 and c[0] == 0 and c[1] == 0:
            alpha = beta = gamma = math.pi / 2
        else:
            alpha = math.acos(np.clip(np.dot(b, c) / (lb * lc), -1, 1))
            beta = math.acos(np.clip(np.dot(c, a) / (lc * la), -1, 1))
            gamma = math.acos(np.clip(np.dot(a, b) / (la * lb), -1, 1))

            # if any angle is very close to PI/2, set it to exactly PI/2
            if abs(alpha - math.pi / 2) < 1e-4:
                alpha = math.pi / 2
            if abs(beta - math.pi / 2) < 1e-4:
                beta = math.pi / 2
            if abs(gamma - math.pi / 2) < 1e-4:
                gamma = math.pi / 2

        return np.array([la, lb, lc], dtype=float), np.array([alpha, beta, gamma], dtype=float)



import math
import numpy as np
from mstk.topology.atom import Atom
from mstk.topology.molecule import Molecule
from mstk.topology.topology import Topology
from mstk.chem.element import Element
from mstk.chem.constant import *


class Zmat():
    '''
    Generate Topology from ZMAT file.

    All of the atoms are assumed to be in the same molecule.
    The first line is treated as the name of the molecule.
    The positions will be generated from Z-matrix, with the algorithm copied from fftool of Agilio Padua.
    The first column is treated as atom type.

    Parameters
    ----------
    file : str
    kwargs : dict
        Ignored

    Attributes
    ----------
    topology : Topology

    Examples
    --------
    >>> zmat = Zmat('input.zmat')
    >>> topology = zmat.topology
    '''

    def __init__(self, file, **kwargs):
        self.topology = Topology()
        self._parse(file)

    def _parse(self, file):
        f = open(file)

        mol = Molecule()
        z_atoms = []

        # read molecule name
        line = f.readline()
        while line.strip().startswith('#'):
            line = f.readline()
        mol.name = line.strip()

        # read z-matrix
        line = f.readline()
        while line.strip().startswith('#') or line.strip() == '':
            line = f.readline()

        words = line.strip().split()
        if len(words) > 1:  # there can be line numbers
            shift = 1
        else:
            shift = 0

        _read_variable = False
        while line and not line.strip().lower().startswith('var'):
            words = line.strip().split()
            if len(words) == 0:
                break
            atom = Atom()
            atom.type = words[shift]
            element = Element.guess_from_atom_type(atom.type)
            atom.symbol = element.symbol
            atom.mass = element.mass
            atom.name = atom.symbol + str(mol.n_atom + 1)
            mol.add_atom(atom)

            ir = ia = id = 0
            r = a = d = 0.0
            var_r = var_a = var_d = ''
            if (len(words) - shift) > 1:
                ir = int(words[shift + 1]) - 1
                mol.add_bond(mol.atoms[ir], atom)

                if words[shift + 2][0].isalpha():
                    var_r = words[shift + 2]
                    _read_variable = True
                else:
                    r = float(words[shift + 2]) / 10  # convert from A to nm
                if (len(words) - shift) > 3:
                    ia = int(words[shift + 3]) - 1
                    if words[shift + 4][0].isalpha():
                        var_a = words[shift + 4]
                        _read_variable = True
                    else:
                        a = float(words[shift + 4])
                    if (len(words) - shift) > 5:
                        id = int(words[shift + 5]) - 1
                        if words[shift + 6][0].isalpha():
                            var_d = words[shift + 6]
                            _read_variable = True
                        else:
                            d = float(words[shift + 6])
            za = {'name': type,
                  'ir'  : ir, 'var_r': var_r, 'r': r,
                  'ia'  : ia, 'var_a': var_a, 'a': a,
                  'id'  : id, 'var_d': var_d, 'd': d}
            z_atoms.append(za)
            line = f.readline()

        # read variables
        if _read_variable:
            if line.strip().lower().startswith('var') or line.strip() == '':
                line = f.readline()
            while line:
                words = line.strip().split('=')
                if len(words) < 2:
                    break
                key = words[0].strip()
                val = float(words[1])
                for za in z_atoms:
                    if za['var_r'] == key:
                        za['r'] = val / 10  # covert from A to nm
                    if za['var_a'] == key:
                        za['a'] = val
                    if za['var_d'] == key:
                        za['d'] = val
                line = f.readline()

        # read extra bonds
        while line:
            if line.strip().startswith('#') or line.strip() == '':
                line = f.readline()
                continue
            words = line.strip().split()
            if words[0] == 'connect':
                ii = int(words[1]) - 1
                ij = int(words[2]) - 1
                mol.add_bond(mol.atoms[ii], mol.atoms[ij])
            line = f.readline()

        f.close()

        if mol.n_atom >= 1:
            mol.atoms[0].position = [0, 0, 0]

        if mol.n_atom >= 2:
            mol.atoms[1].position = [z_atoms[1]['r'], 0, 0]

        # third atom at distance r from ir forms angle a 3-ir-ia in plane xy
        if mol.n_atom >= 3:
            r = z_atoms[2]['r']
            ir = z_atoms[2]['ir']
            ang = z_atoms[2]['a'] * DEG2RAD
            ia = z_atoms[2]['ia']

            # for this construction, the new atom is at point (x, y), atom
            # ir is at point (xr, yr) and atom ia is at point (xa, ya).
            # Theta is the angle between the vector joining ir to ia and
            # the x-axis, a' (= theta - a) is is the angle between r and
            # the x-axis. x = xa + r cos a', y = ya + r sin a'.  From the
            # dot product of a unitary vector along x with the vector from
            # ir to ia, theta can be calculated: cos theta = (xa - xr) /
            # sqrt((xa - xr)^2 + (ya - yr)^2).  If atom ia is in third or
            # forth quadrant relative to atom ir, ya - yr < 0, then theta
            # = 2 pi - theta. */
            dx, dy, dz = mol.atoms[ia].position - mol.atoms[ir].position
            theta = math.acos(np.clip(dx / math.sqrt(dx * dx + dy * dy), -1, 1))
            if dy < 0.0:
                theta = 2 * math.pi - theta
            ang = theta - ang
            mol.atoms[2].position = [mol.atoms[ir].position[0] + r * math.cos(ang),
                                     mol.atoms[ir].position[1] + r * math.sin(ang),
                                     0.0]

        # nth atom at distance r from atom ir forms angle a at 3-ir-ia
        # and dihedral angle between planes 3-ir-ia and ir-ia-id
        if mol.n_atom >= 4:
            for i in range(3, mol.n_atom):
                r = z_atoms[i]['r']
                ir = z_atoms[i]['ir']
                ang = z_atoms[i]['a'] * DEG2RAD
                ia = z_atoms[i]['ia']
                dih = z_atoms[i]['d'] * DEG2RAD
                id = z_atoms[i]['id']

                # for this construction the new atom is at point A, atom ir is
                # at B, atom ia at C and atom id at D.  Point a is the
                # projection of A onto the plane BCD.  Point b is the
                # projection of A along the direction BC (the line defining
                # the dihedral angle between planes ABC and BCD). n = CD x BC
                # / |CD x BC| is the unit vector normal to the plane BCD. m =
                # BC x n / |BC x n| is the unit vector on the plane BCD normal
                # to the direction BC.
                #
                #                               .'A
                #                 ------------.' /.-----------------
                #                /           b /  .               /
                #               /           ./    .              /
                #              /           B......a      ^      /
                #             /           /              |n    /
                #            /           /                    /
                #           /           C                    /
                #          /             \                  /
                #         /               \                /
                #        /plane BCD        D              /
                #       ----------------------------------
                #
                #                    A              C------B...b
                #                   /.             /        .  .
                #                  / .            /    |m    . .
                #                 /  .           /     V      ..
                #         C------B...b          D              a
                #

                BA = r
                vB = np.array(mol.atoms[ir].position)
                vC = np.array(mol.atoms[ia].position)
                vD = np.array(mol.atoms[id].position)

                vBC = vC - vB
                vCD = vD - vC

                BC = math.sqrt(vBC.dot(vBC))
                bB = BA * math.cos(ang)
                bA = BA * math.sin(ang)
                aA = bA * math.sin(dih)
                ba = bA * math.cos(dih)

                vb = vC - vBC * ((BC - bB) / BC)
                vn = np.cross(vCD, vBC)
                vn = vn / math.sqrt(vn.dot(vn))
                vm = np.cross(vBC, vn)
                vm = vm / math.sqrt(vm.dot(vm))
                va = vb + vm * ba
                vA = va + vn * aA

                mol.atoms[i].position = vA

        mol.generate_angle_dihedral_improper()

        self.topology.update_molecules([mol])


Topology.register_format('.zmat', Zmat)



import numpy as np
from pandas import Series
from .canny import canny1d

__all__ = [
    'check_vle_density',
    'check_interface',
    'N_vaporize_condense'
]


class Ptre():
    _PATTERN_A = 'a'
    _PATTERN_B = 'b'
    _PATTERN_C = 'c'
    _PATTERN_D = 'd'

    @staticmethod
    def test_data(data, interval, nms_t=11, low_t=0.3, hpw_t=1.0, debug=False):
        """ Test data and entry of other tests.
            Arguments:
                nms_t, low_t is written in canny.py
        """
        peaks = canny1d(interval, data, nms_t, low_t, debug=debug)
        ave = sum([abs(p[1]) for p in peaks]) / len(peaks)

        if debug:
            print('ave peak:', ave)

        if len(peaks) == 0:
            result = 'b'
        elif len(peaks) == 1:
            # if peaks[0][2] * 2 * hpw_t < len(data):
            #     result = 'a'
            # else:
            #     result = 'c'
            result = 'c'
        elif len(peaks) == 2:
            if not Ptre.test_updown(peaks):
                result = 'Error: Two peaks with same direction.'
            if Ptre.test_hpw(peaks, len(data) * interval, hpw_t, debug):
                result = 'a'
            else:
                result = 'c'
        else:
            if ave > 0.017:  # this value change with gauss_sigma. be careful when modify either
                result = 'd'
            else:
                result = 'b'

        return result, peaks

    # half peak width test
    @staticmethod
    def test_hpw(peaks, length, threshold, debug=False):

        for i in range(0, len(peaks) - 1):
            hpw_ratio = (peaks[i + 1][0] - peaks[i][0]) / (peaks[i][2] + peaks[i + 1][2])
            if debug:
                print('hpw ratio:', i, hpw_ratio)
            if hpw_ratio < threshold:
                return False
        hpw_ratio = (peaks[0][0] + length - peaks[-1][0]) / (peaks[-1][2] + peaks[0][2])
        if debug:
            print('hpw ratio:', -1, hpw_ratio)
        if hpw_ratio < threshold:
            return False
        return True

    @staticmethod
    def test_updown(peaks):
        for i in range(len(peaks) - 1):
            if peaks[i][0] * peaks[i + 1][0] > 0:
                return False
        return True


def check_vle_density(series):
    '''
    Check whether or not a density profile represents a vapor-liquid interface with canny edge method

    Parameters
    ----------
    series : Series
        The density series along z axis. The index of the series is z coordinates

    Returns
    -------
    is_interface : bool
        Whether or not there is a vapor-liquid interface
    is_center_gas : bool
        Whether or not the center region is gas phase
    nodes : list of float
        The location of interfaces in z axis

    '''
    interval = series.index[1] - series.index[0]

    result, peaks = Ptre.test_data(series, interval)  # the peaks starts from 0, not original index
    peaks._sort_hill(key=lambda x: x[0])

    if result != Ptre._PATTERN_A:
        return False, None, None

    nodes = [p[0] + series.index[0] for p in peaks]
    return True, peaks[0][1] > 0, nodes


def N_vaporize_condense(phases):
    '''
    Check how many times one molecules have vaporized or condensed

    Parameters
    ----------
    phases : list of str
        The time evolution of the phase one molecule stayed in.
        The elements in this list can only be 'l', 'i' or 'g', which means liquid, interface and gas phase.

    Returns
    -------
    N_vapor: int
        Times of vaporization (transit from liquid to gas phase)
    N_condense: int
        Times of condensation (transit from gas to liquid phase)

    '''
    N_vapor = N_condense = 0
    phases = [i for i in phases if i != 'i']
    while len(phases) > 1:
        if phases[0] == 'l' and phases[1] == 'g':
            N_vapor += 1
        elif phases[0] == 'g' and phases[1] == 'l':
            N_condense += 1
        phases.pop(0)

    return N_vapor, N_condense


def check_interface(series: Series, debug=False) -> (bool, list):
    interval = series.index[1] - series.index[0]

    result, peaks = Ptre.test_data(series, interval, debug=debug)

    if debug:
        try:
            import pylab
        except:
            print('matplotlib not found, cannot debug')
        else:
            t = np.array(series.index) - series.index[0]
            pylab.plot(t, series)
            if result in [Ptre._PATTERN_A, Ptre._PATTERN_D]:
                pylab.vlines([p[0] for p in peaks], 0, 1, colors='red')
            pylab.show()

            for i in range(0, len(peaks)):
                if peaks[i][1] < 0:
                    state = 'INCREASING'
                else:
                    state = 'DECREASING'
                print(state, peaks[i][0])

    if result == Ptre._PATTERN_A:
        return True, peaks
    else:
        return False, peaks



import itertools
from mstk import logger
from mstk.forcefield.forcefield import ForceField
from mstk.forcefield.ffterm import *


class Zff:
    '''
    Parse ForceField from ZFF file.

    ZFF is the default plain text force field format in `mstk`.
    Comparing to ZFP, ZFF is more human-friendly but less program-friendly.

    In ZFF file, there is no leading 1/2 for harmonic energy terms.
    The length is in unit of nm, and angle is in unit of degree.
    The energies for harmonic bond, harmonic angle and vdW are in unit of kJ/mol/nm^2, kJ/mol/rad^2 and kJ/mol.
    The two values for LJ interaction are epsilon and sigma.

    A force field term describing one topology element should only appear once.
    If there are duplicated terms, an Exception will be raised.

    There are three types of comments:
    1. Lines start with '#' are completely ignored by the parser.
    2. Lines start with '*' are comments for the force field. They will be stored in the ForceField object as a list of str.
       Each such line will be an element of comments.
    3. '#' in the middle of line is separator for parameters and comments. These comments will be stored in the FFTerm object as a list of str.
       The comments in one line can contain several elements, separated by ';'

    Parameters
    ----------
    file : str

    Attributes
    ----------
    forcefield : ForceField
    '''

    def __init__(self, file):
        self.forcefield = ForceField()
        self._parse(file)

    def _parse(self, file):
        with open(file) as f:
            lines = f.read().splitlines()

        ff = self.forcefield
        _duplicated = []
        for line in lines:
            if line.startswith('*'):
                ff.comments.append(line.lstrip('*').strip())
                continue
            parts = line.split('#')
            str_term = parts[0].strip()
            if str_term == '':
                continue
            if str_term.startswith('Setting'):
                words = str_term.split()
                key, str_val = words[1:]
                if key not in ForceField.SETTING_ATTRS:
                    raise Exception('Invalid setting: ' + key)
                setattr(ff, key, ForceField.SETTING_ATTRS[key](str_val))
                continue

            try:
                term = FFTermFactory.create_from_zff(str_term)
            except:
                raise Exception('Invalid force field line: ' + str_term)

            if len(parts) > 1:
                for comment in parts[1].split(';'):
                    c = comment.strip()
                    if c:
                        term.comments.append(c)

            try:
                ff.add_term(term)
            except:
                _duplicated.append(term)

        if _duplicated:
            msg = f'{len(_duplicated)} duplicated terms from {file}'
            for term in _duplicated:
                msg += f'\n        {term}'
            logger.error(msg)
            raise Exception('Duplicated terms in ZFF file')

    @staticmethod
    def save_to(ff, file):
        '''
        Save ForceField to a ZFF file

        Parameters
        ----------
        ff : ForceField
        file : str
        '''
        line = ''
        if ff.comments:
            for comment in ff.comments:
                line += f'* {comment}\n'
            line += '\n'

        for attr in ForceField.SETTING_ATTRS:
            line += '%-16s %-16s %s\n' % ('Setting', attr, getattr(ff, attr))
        line += '\n'

        _header_printed = set()
        for term in itertools.chain(ff.atom_types.values(),
                                    ff.virtual_site_terms.values(),
                                    ff.qinc_terms.values(),
                                    ff.vdw_terms.values(),
                                    ff.pairwise_vdw_terms.values(),
                                    ff.bond_terms.values(),
                                    ff.angle_terms.values(),
                                    ff.dihedral_terms.values(),
                                    ff.improper_terms.values(),
                                    ff.polarizable_terms.values()):
            alias = type(term).get_alias()
            if alias not in _header_printed:
                line += term.to_zff_header() + '\n'
                _header_printed.add(alias)
            line += term.to_zff()
            if term.comments:
                line += '  # ' + '; '.join(term.comments)
            line += '\n'

        with open(file, 'wb') as f:
            f.write(line.encode())


ForceField.registor_format('.zff', Zff)



import numpy as np
from ..chem import constant


def ew_dot(vec1, vec2):
    '''
    Element-wise dot product
    '''
    return np.sum(vec1 * vec2, axis=1)


class HarmonicBondKernel():
    '''
    E = k (r-r0)^2

    Parameters
    ----------
    positions : list of list of float
    indexes : list of list of int
    parameters : list of list of float
    '''

    def __init__(self, positions, indexes, parameters):
        self.positions = np.array(positions, dtype=np.float64)
        self.indexes = np.array(indexes, dtype=int)
        self.a1 = self.indexes[:, 0]
        self.a2 = self.indexes[:, 1]
        self.parameters = np.array(parameters, dtype=np.float64)
        self.r0 = self.parameters[:, 0]
        self.k = self.parameters[:, 1]

    def evaluate(self):
        delta = self.positions[self.a2] - self.positions[self.a1]
        rsq = ew_dot(delta, delta)
        r = np.sqrt(rsq)
        energy = self.k * (r - self.r0) ** 2

        forces_a1 = (2 * self.k * (r - self.r0) / r)[:, np.newaxis] * delta

        forces = np.zeros(self.positions.shape, dtype=np.float64)
        # TODO vectorize this loop
        for i in range(self.positions.shape[0]):
            forces[i] += forces_a1[self.a1 == i].sum(axis=0)
            forces[i] -= forces_a1[self.a2 == i].sum(axis=0)

        return r, energy, forces


class HarmonicAngleKernel():
    '''
    E = k (theta-theta0)^2

    '''

    def __init__(self, positions, indexes, parameters):
        self.positions = np.array(positions, dtype=np.float64)
        self.indexes = np.array(indexes, dtype=int)
        self.a1 = self.indexes[:, 0]
        self.a2 = self.indexes[:, 1]
        self.a3 = self.indexes[:, 2]
        self.parameters = np.array(parameters, dtype=np.float64)
        self.theta0 = self.parameters[:, 0]
        self.k = self.parameters[:, 1]

    def evaluate(self):
        vec1 = self.positions[self.a1] - self.positions[self.a2]
        vec2 = self.positions[self.a3] - self.positions[self.a2]
        r1 = np.sqrt(ew_dot(vec1, vec1))
        r2 = np.sqrt(ew_dot(vec2, vec2))
        cos = ew_dot(vec1, vec2) / r1 / r2
        np.clip(cos, -1, 1, out=cos)
        theta = np.arccos(cos)
        energy = self.k * (theta - self.theta0) ** 2

        sin = np.sqrt(1 - cos * cos)
        sin[sin < 1E-4] = 1E-4

        factor = 2 * self.k * (theta - self.theta0) / sin
        c11 = -factor * cos / r1 / r1
        c12 = factor / r1 / r2
        c31 = -factor * cos / r2 / r2

        forces_a1 = c11[:, np.newaxis] * vec1 + c12[:, np.newaxis] * vec2
        forces_a3 = c31[:, np.newaxis] * vec2 + c12[:, np.newaxis] * vec1

        forces = np.zeros(self.positions.shape, dtype=np.float64)
        # TODO vectorize this loop
        for i in range(self.positions.shape[0]):
            forces[i] += forces_a1[self.a1 == i].sum(axis=0)
            forces[i] += forces_a3[self.a3 == i].sum(axis=0)
            forces[i] -= forces_a1[self.a2 == i].sum(axis=0) + forces_a3[self.a2 == i].sum(axis=0)

        return theta, energy, forces


class OplsTorsionKernel():
    '''
    E = k1 (1+cos(phi)) + k2 (1-cos(2 phi)) + k3 (1+cos(3 phi)) + k4 (1-cos(4 phi))

    '''

    def __init__(self, positions, indexes, parameters):
        self.positions = np.array(positions, dtype=np.float64)
        self.indexes = np.array(indexes, dtype=int)
        self.a1 = self.indexes[:, 0]
        self.a2 = self.indexes[:, 1]
        self.a3 = self.indexes[:, 2]
        self.a4 = self.indexes[:, 3]
        self.parameters = np.array(parameters, dtype=np.float64)
        self.k1 = self.parameters[:, 0]
        self.k2 = self.parameters[:, 1]
        self.k3 = self.parameters[:, 2]
        self.k4 = self.parameters[:, 3]

    def evaluate(self):
        vec1 = self.positions[self.a1] - self.positions[self.a2]
        vec2 = self.positions[self.a3] - self.positions[self.a2]
        vec3 = self.positions[self.a3] - self.positions[self.a4]
        n1 = np.cross(vec1, vec2)
        n2 = np.cross(vec2, vec3)
        rsq_n1 = ew_dot(n1, n1)
        rsq_n2 = ew_dot(n2, n2)
        cos = ew_dot(n1, n2) / np.sqrt(rsq_n1 * rsq_n2)
        np.clip(cos, -1, 1, out=cos)
        sign = np.ones(vec1.shape[0])
        sign[ew_dot(vec1, n2) < 0] = -1
        phi = sign * np.arccos(cos)

        c0 = self.k1 + 2 * self.k2 + self.k3
        c1 = self.k1 - 3 * self.k3
        c2 = -2 * self.k2 + 8 * self.k4
        c3 = 4 * self.k3
        c4 = -8 * self.k4
        cos2 = cos * cos
        cos3 = cos2 * cos
        cos4 = cos3 * cos
        energy = c0 + c1 * cos + c2 * cos2 + c3 * cos3 + c4 * cos4

        sin = np.sin(phi)
        ### (-dE / d phi)
        factor = (c1 * sin
                  + 2 * c2 * cos * sin
                  + 3 * c3 * cos2 * sin
                  + 4 * c4 * cos3 * sin)

        ### (d phi / d ri) copied from OpenMM
        rsq2 = ew_dot(vec2, vec2)
        r2 = np.sqrt(rsq2)

        factor1 = factor * r2 / rsq_n1
        factor4 = -factor * r2 / rsq_n2

        forces_a1 = factor1[:, np.newaxis] * n1
        forces_a4 = factor4[:, np.newaxis] * n2

        factor2 = ew_dot(vec1, vec2) / rsq2
        factor3 = ew_dot(vec3, vec2) / rsq2

        s = factor2[:, np.newaxis] * forces_a1 - factor3[:, np.newaxis] * forces_a4

        forces_a2 = -forces_a1 + s
        forces_a3 = -forces_a4 - s
        ###

        forces = np.zeros(self.positions.shape, dtype=np.float64)
        # TODO vectorize this loop
        for i in range(self.positions.shape[0]):
            forces[i] += forces_a1[self.a1 == i].sum(axis=0)
            forces[i] += forces_a2[self.a2 == i].sum(axis=0)
            forces[i] += forces_a3[self.a3 == i].sum(axis=0)
            forces[i] += forces_a4[self.a4 == i].sum(axis=0)

        return phi, energy, forces


class HarmonicTorsionKernel():
    '''
    E = k (phi-phi0)^2

    '''

    def __init__(self, positions, indexes, parameters):
        self.positions = np.array(positions, dtype=np.float64)
        self.indexes = np.array(indexes, dtype=int)
        self.a1 = self.indexes[:, 0]
        self.a2 = self.indexes[:, 1]
        self.a3 = self.indexes[:, 2]
        self.a4 = self.indexes[:, 3]
        self.parameters = np.array(parameters, dtype=np.float64)
        self.phi0 = self.parameters[:, 0]
        self.k = self.parameters[:, 1]

    def evaluate(self):
        vec1 = self.positions[self.a1] - self.positions[self.a2]
        vec2 = self.positions[self.a3] - self.positions[self.a2]
        vec3 = self.positions[self.a3] - self.positions[self.a4]
        n1 = np.cross(vec1, vec2)
        n2 = np.cross(vec2, vec3)
        rsq_n1 = ew_dot(n1, n1)
        rsq_n2 = ew_dot(n2, n2)
        cos = ew_dot(n1, n2) / np.sqrt(rsq_n1 * rsq_n2)
        np.clip(cos, -1, 1, out=cos)
        sign = np.ones(vec1.shape[0])
        sign[ew_dot(vec1, n2) < 0] = -1
        phi = sign * np.arccos(cos)

        d_phi = phi - self.phi0
        d_phi_abs = np.abs(d_phi)
        d_phi_2pi = 2 * np.pi - d_phi_abs
        d_phi_min = np.min([d_phi_abs, d_phi_2pi], axis=0)
        energy = self.k * d_phi_min ** 2

        ### (-dE / d phi)
        factor = - 2 * self.k * d_phi_min

        sign.fill(1)
        sign[d_phi_2pi < d_phi_abs] = -1
        factor *= sign

        sign.fill(1)
        sign[d_phi < 0] = -1
        factor *= sign

        ### (d phi / d ri) copied from OpenMM
        rsq2 = ew_dot(vec2, vec2)
        r2 = np.sqrt(rsq2)

        factor1 = factor * r2 / rsq_n1
        factor4 = -factor * r2 / rsq_n2

        forces_a1 = factor1[:, np.newaxis] * n1
        forces_a4 = factor4[:, np.newaxis] * n2

        factor2 = ew_dot(vec1, vec2) / rsq2
        factor3 = ew_dot(vec3, vec2) / rsq2

        s = factor2[:, np.newaxis] * forces_a1 - factor3[:, np.newaxis] * forces_a4

        forces_a2 = -forces_a1 + s
        forces_a3 = -forces_a4 - s
        ###

        forces = np.zeros(self.positions.shape, dtype=np.float64)
        # TODO vectorize this loop
        for i in range(self.positions.shape[0]):
            forces[i] += forces_a1[self.a1 == i].sum(axis=0)
            forces[i] += forces_a2[self.a2 == i].sum(axis=0)
            forces[i] += forces_a3[self.a3 == i].sum(axis=0)
            forces[i] += forces_a4[self.a4 == i].sum(axis=0)

        return phi, energy, forces


class ConstrainedTorsionKernel():
    '''
    E = k (1-cos(phi-phi0))

    '''

    def __init__(self, positions, indexes, parameters):
        self.positions = np.array(positions, dtype=np.float64)
        self.indexes = np.array(indexes, dtype=int)
        self.a1 = self.indexes[:, 0]
        self.a2 = self.indexes[:, 1]
        self.a3 = self.indexes[:, 2]
        self.a4 = self.indexes[:, 3]
        self.parameters = np.array(parameters, dtype=np.float64)
        self.phi0 = self.parameters[:, 0]
        self.k = self.parameters[:, 1]

    def evaluate(self):
        vec1 = self.positions[self.a1] - self.positions[self.a2]
        vec2 = self.positions[self.a3] - self.positions[self.a2]
        vec3 = self.positions[self.a3] - self.positions[self.a4]
        n1 = np.cross(vec1, vec2)
        n2 = np.cross(vec2, vec3)
        rsq_n1 = ew_dot(n1, n1)
        rsq_n2 = ew_dot(n2, n2)
        cos = ew_dot(n1, n2) / np.sqrt(rsq_n1 * rsq_n2)
        np.clip(cos, -1, 1, out=cos)
        sign = np.ones(vec1.shape[0])
        sign[ew_dot(vec1, n2) < 0] = -1
        phi = sign * np.arccos(cos)

        d_phi = phi - self.phi0
        energy = self.k * (1 - np.cos(d_phi))

        ### (-dE / d phi)
        factor = -self.k * np.sin(d_phi)

        ### (d phi / d ri) copied from OpenMM
        rsq2 = ew_dot(vec2, vec2)
        r2 = np.sqrt(rsq2)

        factor1 = factor * r2 / rsq_n1
        factor4 = -factor * r2 / rsq_n2

        forces_a1 = factor1[:, np.newaxis] * n1
        forces_a4 = factor4[:, np.newaxis] * n2

        factor2 = ew_dot(vec1, vec2) / rsq2
        factor3 = ew_dot(vec3, vec2) / rsq2

        s = factor2[:, np.newaxis] * forces_a1 - factor3[:, np.newaxis] * forces_a4

        forces_a2 = -forces_a1 + s
        forces_a3 = -forces_a4 - s
        ###

        forces = np.zeros(self.positions.shape, dtype=np.float64)
        # TODO vectorize this loop
        for i in range(self.positions.shape[0]):
            forces[i] += forces_a1[self.a1 == i].sum(axis=0)
            forces[i] += forces_a2[self.a2 == i].sum(axis=0)
            forces[i] += forces_a3[self.a3 == i].sum(axis=0)
            forces[i] += forces_a4[self.a4 == i].sum(axis=0)

        return phi, energy, forces


class NonbondedKernel():
    '''
    E = 4 * eps*((sig/r)^12 - (sig/r)^6) + 138.935455 * qq/r
    '''

    def __init__(self, positions, indexes, parameters):
        self.positions = np.array(positions, dtype=np.float64)
        self.indexes = np.array(indexes, dtype=int)
        self.a1 = self.indexes[:, 0]
        self.a2 = self.indexes[:, 1]
        self.parameters = np.array(parameters, dtype=np.float64)
        self.c12 = 4 * self.parameters[:, 0] * self.parameters[:, 1] ** 12
        self.c6 = 4 * self.parameters[:, 0] * self.parameters[:, 1] ** 6
        self.qq = self.parameters[:, 2]
        self.qqconv = constant.ONE_4PI_EPS0

    def evaluate(self):
        delta = self.positions[self.a2] - self.positions[self.a1]
        rsq = ew_dot(delta, delta)
        r = np.sqrt(rsq)
        r6 = r ** 6
        r12 = r6 * r6
        energy = self.c12 / r12 - self.c6 / r6 + self.qqconv * self.qq / r

        forces_a1 = -(12 * self.c12 / r12 / rsq - 6 * self.c6 / r6 / rsq
                      + self.qqconv * self.qq / rsq / r) \
            [:, np.newaxis] * delta

        forces = np.zeros(self.positions.shape, dtype=np.float64)
        # TODO vectorize this loop
        for i in range(self.positions.shape[0]):
            forces[i] += forces_a1[self.a1 == i].sum(axis=0)
            forces[i] -= forces_a1[self.a2 == i].sum(axis=0)

        return r, energy, forces
